import math

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import EngFormatter, LogFormatterSciNotation


def plotLine(c0, c1, ax, **kwargs):
    t = np.linspace(0, 1, 11)
    c = (c1 - c0) * t + c0
    ax.plot(c.real, c.imag, **kwargs)


def plotCircle(c0, r, ax, **kwargs):
    t = np.linspace(0, 1, 1001) * 2 * np.pi
    s = c0 + r * np.exp(1j * t)
    ax.plot(s.real, s.imag, **kwargs)


def plotEllipse(c0, a, b, phi, ax, **kwargs):
    t = np.linspace(0, 1, 1001) * 2 * np.pi
    c = np.exp(1j * t)
    s = c0 + (c.real * a + 1j * c.imag * b) * np.exp(1j * phi)
    ax.plot(s.real, s.imag, **kwargs)


def plotDistribution(s0,
                     s1,
                     fig=None,
                     axes=None,
                     info=None,
                     hotThresh=10000,
                     logy=False):
    from waveforms.math.fit import get_threshold_info, mult_gaussian_pdf

    if info is None:
        info = get_threshold_info(s0, s1)
    else:
        info = get_threshold_info(s0, s1, info['threshold'], info['phi'])
    thr, phi = info['threshold'], info['phi']
    # visibility, p0, p1 = info['visibility']
    # print(
    #     f"thr={thr:.6f}, phi={phi:.6f}, visibility={visibility:.3f}, {p0}, {1-p1}"
    # )

    if axes is not None:
        ax1, ax2 = axes
    else:
        if fig is None:
            fig = plt.figure()
        ax1 = fig.add_subplot(121)
        ax2 = fig.add_subplot(122)

    if (len(s0) + len(s1)) < hotThresh:
        ax1.plot(np.real(s0), np.imag(s0), '.', alpha=0.2)
        ax1.plot(np.real(s1), np.imag(s1), '.', alpha=0.2)
    else:
        _, *bins = np.histogram2d(np.real(np.hstack([s0, s1])),
                                  np.imag(np.hstack([s0, s1])),
                                  bins=50)

        H0, *_ = np.histogram2d(np.real(s0),
                                np.imag(s0),
                                bins=bins,
                                density=True)
        H1, *_ = np.histogram2d(np.real(s1),
                                np.imag(s1),
                                bins=bins,
                                density=True)
        vlim = max(np.max(np.abs(H0)), np.max(np.abs(H1)))

        ax1.imshow(H1.T - H0.T,
                   alpha=(np.fmax(H0.T, H1.T) / vlim).clip(0, 1),
                   interpolation='nearest',
                   origin='lower',
                   cmap='coolwarm',
                   vmin=-vlim,
                   vmax=vlim,
                   extent=(bins[0][0], bins[0][-1], bins[1][0], bins[1][-1]))

    ax1.axis('equal')
    ax1.set_xticks([])
    ax1.set_yticks([])
    for s in ax1.spines.values():
        s.set_visible(False)

    # c0, c1 = info['center']
    # a0, b0, a1, b1 = info['std']
    params = info['params']
    r0, i0, r1, i1 = params[0][0], params[1][0], params[0][1], params[1][1]
    a0, b0, a1, b1 = params[0][2], params[1][2], params[0][3], params[1][3]
    c0 = (r0 + 1j * i0) * np.exp(1j * phi)
    c1 = (r1 + 1j * i1) * np.exp(1j * phi)
    phi0 = phi + params[0][6]
    phi1 = phi + params[1][6]
    plotEllipse(c0, 2 * a0, 2 * b0, phi0, ax1)
    plotEllipse(c1, 2 * a1, 2 * b1, phi1, ax1)

    im0, im1 = info['idle']
    lim = min(im0.min(), im1.min()), max(im0.max(), im1.max())
    t = (np.linspace(lim[0], lim[1], 3) + 1j * thr) * np.exp(-1j * phi)
    ax1.plot(t.imag, t.real, 'k--')

    ax1.plot(np.real(c0), np.imag(c0), 'o', color='C3')
    ax1.plot(np.real(c1), np.imag(c1), 'o', color='C4')

    re0, re1 = info['signal']
    x, a, b, c = info['cdf']

    xrange = (min(re0.min(), re1.min()), max(re0.max(), re1.max()))

    n0, bins0, *_ = ax2.hist(re0, bins=80, range=xrange, alpha=0.5)
    n1, bins1, *_ = ax2.hist(re1, bins=80, range=xrange, alpha=0.5)

    x_range = np.linspace(x.min(), x.max(), 1001)
    *_, cov0, cov1 = info['std']
    ax2.plot(
        x_range,
        np.sum(n0) * (bins0[1] - bins0[0]) *
        mult_gaussian_pdf(x_range, [r0, r1], [
            np.sqrt(cov0[0, 0]), np.sqrt(cov1[0, 0])
        ], [params[0][4], 1 - params[0][4]]))
    ax2.plot(
        x_range,
        np.sum(n1) * (bins1[1] - bins1[0]) *
        mult_gaussian_pdf(x_range, [r0, r1], [
            np.sqrt(cov0[0, 0]), np.sqrt(cov1[0, 0])
        ], [params[0][5], 1 - params[0][5]]))
    ax2.set_ylabel('Count')
    ax2.set_xlabel('Projection Axes')
    if logy:
        ax2.set_yscale('log')
        ax2.set_ylim(0.1, max(np.sum(n0), np.sum(n1)))

    ax3 = ax2.twinx()
    ax3.plot(x, a, '--', lw=1, color='C0')
    ax3.plot(x, b, '--', lw=1, color='C1')
    ax3.plot(x, c, 'k--', alpha=0.5, lw=1)
    ax3.set_ylim(0, 1.1)
    ax3.vlines(thr, 0, 1.1, 'k', alpha=0.5)
    ax3.set_ylabel('Probility')

    return info


ALLXYSeq = [('I', 'I'), ('X', 'X'), ('Y', 'Y'), ('X', 'Y'), ('Y', 'X'),
            ('X/2', 'I'), ('Y/2', 'I'), ('X/2', 'Y/2'), ('Y/2', 'X/2'),
            ('X/2', 'Y'), ('Y/2', 'X'), ('X', 'Y/2'), ('Y', 'X/2'),
            ('X/2', 'X'), ('X', 'X/2'), ('Y/2', 'Y'), ('Y', 'Y/2'), ('X', 'I'),
            ('Y', 'I'), ('X/2', 'X/2'), ('Y/2', 'Y/2')]


def plotALLXY(data, ax=None):
    assert len(data) % len(ALLXYSeq) == 0

    if ax is None:
        ax = plt.gca()

    ax.plot(np.array(data), 'o-')
    repeat = len(data) // len(ALLXYSeq)
    ax.set_xticks(np.arange(len(ALLXYSeq)) * repeat + 0.5 * (repeat - 1))
    ax.set_xticklabels([','.join(seq) for seq in ALLXYSeq], rotation=60)
    ax.grid(which='major')


def _get_log_ticks(x):
    log10x = np.log10(x)

    major_ticks = np.array(
        range(math.floor(log10x[0]) - 1,
              math.ceil(log10x[-1]) + 1))
    minor_ticks = np.hstack([
        np.log10(np.linspace(2, 10, 9, endpoint=False)) + x
        for x in major_ticks
    ])

    major_ticks = major_ticks[(major_ticks >= log10x[0]) *
                              (major_ticks <= log10x[-1])]
    minor_ticks = minor_ticks[(minor_ticks >= log10x[0]) *
                              (minor_ticks <= log10x[-1])]

    return log10x, major_ticks, minor_ticks


class MyLogFormatter(EngFormatter):

    def format_ticks(self, values):
        if self.unit is None or self.unit == '':
            fmt = LogFormatterSciNotation()
            return [f"${fmt.format_data(10.0**x)}$" for x in values]
        else:
            return super().format_ticks(values)

    def format_eng(self, x):
        if self.unit is None or self.unit == '':
            self.unit = ''
            return f"{10.0**x:g}"
        else:
            return super().format_eng(10.0**x)


def imshow_logx(x, y, z, x_unit=None, ax=None, **kwargs):
    if ax is None:
        ax = plt.gca()

    log10x, major_ticks, minor_ticks = _get_log_ticks(x)

    dlogx, dy = log10x[1] - log10x[0], y[1] - y[0]
    extent = (log10x[0] - dlogx / 2, log10x[-1] + dlogx / 2, y[0] - dy / 2,
              y[-1] + dy / 2)

    img = ax.imshow(z, extent=extent, **kwargs)

    ax.set_xticks(major_ticks, minor=False)
    ax.xaxis.set_major_formatter(MyLogFormatter(x_unit))
    ax.set_xticks(minor_ticks, minor=True)

    return img


def imshow_logy(x, y, z, y_unit=None, ax=None, **kwargs):
    if ax is None:
        ax = plt.gca()

    log10y, major_ticks, minor_ticks = _get_log_ticks(y)

    dlogy, dx = log10y[1] - log10y[0], x[1] - x[0]
    extent = (x[0] - dx / 2, x[-1] + dx / 2, log10y[0] - dlogy / 2,
              log10y[-1] + dlogy / 2)

    img = ax.imshow(z, extent=extent, **kwargs)

    ax.set_yticks(major_ticks, minor=False)
    ax.yaxis.set_major_formatter(MyLogFormatter(y_unit))
    ax.set_yticks(minor_ticks, minor=True)

    return img


def imshow_loglog(x, y, z, x_unit=None, y_unit=None, ax=None, **kwargs):
    if ax is None:
        ax = plt.gca()

    log10x, x_major_ticks, x_minor_ticks = _get_log_ticks(x)
    log10y, y_major_ticks, y_minor_ticks = _get_log_ticks(y)

    dlogx, dlogy = log10x[1] - log10x[0], log10y[1] - log10y[0]
    extent = (log10x[0] - dlogx / 2, log10x[-1] + dlogx / 2,
              log10y[0] - dlogy / 2, log10y[-1] + dlogy / 2)

    img = ax.imshow(z, extent=extent, **kwargs)

    ax.set_xticks(x_major_ticks, minor=False)
    ax.xaxis.set_major_formatter(MyLogFormatter(x_unit))
    ax.set_xticks(x_minor_ticks, minor=True)

    ax.set_yticks(y_major_ticks, minor=False)
    ax.yaxis.set_major_formatter(MyLogFormatter(y_unit))
    ax.set_yticks(y_minor_ticks, minor=True)

    return img


def autoplot(x,
             y,
             z,
             xlabel='x',
             ylabel='y',
             zlabel='z',
             x_unit='',
             y_unit='',
             z_unit='',
             fig=None,
             ax=None,
             index=None,
             xscale='linear',
             yscale='linear',
             zscale='linear',
             max_lines=3,
             **kwds):
    """
    Plot a 2D array as a line plot or an image.

    Parameters:
        x (array): x values
        y (array): y values
        z (array): z values
        xlabel (str): x label
        ylabel (str): y label
        zlabel (str): z label
        x_unit (str): x unit
        y_unit (str): y unit
        z_unit (str): z unit
        fig (Figure): figure to plot on
        ax (Axes): axes to plot on
        index (int): index of the line to plot
        xscale (str): x scale
        yscale (str): y scale
        zscale (str): z scale
        max_lines (int): maximum number of lines to plot
        **kwds: keyword arguments passed to plot_img or plot_lines
    """
    if ax is not None:
        fig = ax.figure
    if fig is None:
        fig = plt.gcf()
    if ax is None:
        ax = fig.add_subplot(111)

    if len(y) <= max_lines or len(x) <= max_lines or index is not None:
        plot_lines(x,
                   y,
                   z,
                   xlabel,
                   ylabel,
                   zlabel,
                   x_unit=x_unit,
                   y_unit=y_unit,
                   z_unit=z_unit,
                   xscale=xscale,
                   yscale=yscale,
                   zscale=zscale,
                   ax=ax,
                   index=index,
                   **kwds)
    else:
        plot_img(x,
                 y,
                 z,
                 xlabel,
                 ylabel,
                 zlabel,
                 x_unit=x_unit,
                 y_unit=y_unit,
                 z_unit=z_unit,
                 xscale=xscale,
                 yscale=yscale,
                 zscale=zscale,
                 fig=fig,
                 ax=ax,
                 **kwds)


def plot_lines(x,
               y,
               z,
               xlabel,
               ylabel,
               zlabel,
               x_unit,
               y_unit,
               z_unit,
               ax,
               xscale='linear',
               yscale='linear',
               zscale='linear',
               index=None,
               **kwds):
    z = np.asarray(z)
    if len(y) > len(x):
        x, y = y, x
        xlabel, ylabel = ylabel, xlabel
        z = z.T
    if index is not None:
        y = y[index]
        z = z[index, :]

    for i, l in enumerate(y):
        if y_unit:
            label = f"{ylabel}={l:.3} [{y_unit}]"
        else:
            label = f"{ylabel}={l:.3}"
        ax.plot(x, z[i, :], label=label, **kwds)
    ax.legend()
    xlabel = f"{xlabel} [{x_unit}]" if x_unit else xlabel
    zlabel = f"{zlabel} [{z_unit}]" if z_unit else zlabel
    ax.set_xlabel(xlabel)
    ax.set_ylabel(zlabel)
    ax.set_xscale(xscale)
    ax.set_yscale(yscale)


def plot_img(x,
             y,
             z,
             xlabel,
             ylabel,
             zlabel,
             x_unit,
             y_unit,
             z_unit,
             fig,
             ax,
             xscale='linear',
             yscale='linear',
             zscale='linear',
             **kwds):
    kwds.setdefault('origin', 'lower')
    kwds.setdefault('aspect', 'auto')
    kwds.setdefault('interpolation', 'nearest')

    zlabel = f"{zlabel} [{z_unit}]" if z_unit else zlabel

    if (xscale, yscale) == ('linear', 'linear'):
        dx, dy = x[1] - x[0], y[1] - y[0]
        extent = (x[0] - dx / 2, x[-1] + dx / 2, y[0] - dy / 2, y[-1] + dy / 2)
        kwds.setdefault('extent', extent)
        img = ax.imshow(np.asarray(z), **kwds)
        xlabel = f"{xlabel} [{x_unit}]" if x_unit else xlabel
        ylabel = f"{ylabel} [{y_unit}]" if y_unit else ylabel
    elif (xscale, yscale) == ('log', 'linear'):
        ylabel = f"{ylabel} [{y_unit}]" if y_unit else ylabel
        img = imshow_logx(x, y, z, x_unit, ax, **kwds)
    elif (xscale, yscale) == ('linear', 'log'):
        xlabel = f"{xlabel} [{x_unit}]" if x_unit else xlabel
        img = imshow_logy(x, y, z, y_unit, ax, **kwds)
    elif (xscale, yscale) == ('log', 'log'):
        img = imshow_loglog(x, y, z, x_unit, y_unit, ax, **kwds)
    else:
        pass
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    cb = fig.colorbar(img, ax=ax)
    cb.set_label(zlabel)



from waveforms.qlisp.config import (CompileConfigMixin, Config, ConfigObject,
                                    ConfigProxy, Gate, Measure, Qubit,
                                    getConfig, rfUnitary, set_config_factory,
                                    setConfig)

__all__ = ['Config', 'getConfig', 'setConfig']



from ..bayes import bayesian_correction
from .delay import calc_delays, fit_relative_delay
from .readout import (cdf, classify, classify_data, count_state, count_to_diag,
                      fit_readout_distribution, fit_readout_distribution2,
                      gaussian_cdf, gaussian_cdf_inv, gaussian_pdf,
                      gaussian_pdf_2d, get_threshold_info,
                      install_classify_method, mult_gaussian_cdf,
                      mult_gaussian_pdf, readout_distribution,
                      uninstall_classify_method)
from .simple import (complex_amp_to_real, find_cross_point, fit_circle,
                     fit_cosine, fit_cross_point, fit_pole, goodness_of_fit,
                     inv_poly, lin_fit, poly_fit)
from .spectrum import (fit_transmon_spectrum, transmon_spectrum,
                       transmon_spectrum_fast)



import pickle
import re
import warnings
from abc import ABC, abstractmethod
from itertools import chain, permutations
from pathlib import Path
from typing import Any, Optional, Union

import numpy as np
from waveforms.baseconfig import BaseConfig, Trait

from .base import (ABCCompileConfigMixin, ADChannel, AWGChannel, GateConfig,
                    MultADChannel, MultAWGChannel, getConfig,
                    set_config_factory)


class CompileConfigMixin(ABCCompileConfigMixin):
    def _getAWGChannel(self, name, *qubits) -> Union[str, dict]:
        def _getSharedCoupler(qubits):
            s = set(qubits[0]['couplers'])
            for qubit in qubits[1:]:
                s = s & set(qubit['couplers'])
            return s

        qubits = [self.getQubit(q) for q in qubits]

        if name.startswith('readoutLine.'):
            name = name.removeprefix('readoutLine.')
            rl = self.getReadoutLine(qubits[0]['readoutLine'])
            chInfo = rl.query('channels.' + name)
        elif name.startswith('coupler.'):
            name = name.removeprefix('coupler.')
            c = _getSharedCoupler(qubits).pop()
            c = self.getCoupler(c)
            chInfo = c.query('channels.' + name)
        else:
            chInfo = qubits[0].query('channels.' + name)
        if isinstance(chInfo, dict):
            return self._makeMultAWGChannel(chInfo)
        else:
            return self._makeAWGChannel(chInfo)

    def _makeAWGChannel(self, name: str) -> AWGChannel:
        info = {'name': name}
        if re.match(r'.*\.Marker[0-9]$', name):
            ch = self.getChannel(name[:name.rfind('.')])
        else:
            ch = self.getChannel(name)
        info['sampleRate'] = ch.params.sampleRate
        return AWGChannel(**info)

    def _makeMultAWGChannel(self, chInfo: dict) -> MultAWGChannel:
        info = {}
        if 'I' in chInfo:
            info['I'] = self._makeAWGChannel(chInfo['I'])
        if 'Q' in chInfo:
            info['Q'] = self._makeAWGChannel(chInfo['Q'])
        if 'LO' in chInfo:
            lo = self.getChannel(chInfo['LO'])
            info['lo_freq'] = lo.status.frequency
        return MultAWGChannel(**info)

    def _getADChannel(self, qubit) -> Union[str, dict]:
        rl = self.getQubit(qubit).readoutLine
        rl = self.getReadoutLine(rl)
        chInfo = rl.channels.AD

        if isinstance(chInfo, str):
            return self._makeADChannel(chInfo)
        else:
            return self._makeMultADChannel(chInfo)

    def _makeADChannel(self,
                       name: str,
                       trigger: str = '',
                       triggerDelay: float = 0) -> ADChannel:
        info = {'name': name}
        ch = self.getChannel(name)
        info['sampleRate'] = ch.params.sampleRate
        info['triggerDelay'] = triggerDelay
        info['trigger'] = trigger
        return ADChannel(**info)

    def _makeMultADChannel(self, chInfo: dict) -> MultADChannel:
        info = {}
        if 'I' in chInfo:
            info['I'] = self._makeADChannel(chInfo['I'],
                                            chInfo.get('trigger', ''),
                                            chInfo.get('triggerDelay', 0))
        if 'Q' in chInfo:
            info['Q'] = self._makeADChannel(chInfo['Q'],
                                            chInfo.get('trigger', ''),
                                            chInfo.get('triggerDelay', 0))
        if 'IQ' in chInfo:
            info['IQ'] = self._makeADChannel(chInfo['IQ'],
                                             chInfo.get('trigger', ''),
                                             chInfo.get('triggerDelay', 0))
        if 'Ref' in chInfo:
            info['Ref'] = self._makeADChannel(chInfo['Ref'],
                                              chInfo.get('trigger', ''),
                                              chInfo.get('triggerDelay', 0))
        if 'LO' in chInfo:
            lo = self.getChannel(chInfo['LO'])
            info['LO'] = chInfo['LO']
            info['lo_freq'] = lo.status.frequency
            info['lo_power'] = lo.status.power
        return MultADChannel(**info)

    def _getGateConfig(self, name, *qubits, type=None) -> GateConfig:
        try:
            gate = self.getGate(name, *qubits)
        except:
            return GateConfig(name, qubits)
        if type is None:
            type = gate.get('default_type', 'default')
        if type not in gate and 'params' in gate:
            params = gate['params']
        else:
            try:
                params = gate[type]
            except KeyError:
                raise KeyError(f'Gate {name} has no type {type}')
        return GateConfig(name, gate['qubits'], type, params)

    def _getAllQubitLabels(self) -> list[str]:
        return sorted(self['chip']['qubits'].keys(), key=lambda s: int(s[1:]))


class Config(BaseConfig, CompileConfigMixin):
    def __init__(self, path: Optional[Union[str, Path]] = None):
        super().__init__(path)
        if 'station' not in self:
            self.update({
                'station': {
                    'name': "",
                    'instruments': {},
                    'channels': {},
                    'params': {},
                    'status': {},
                    'calibrations': {}
                },
                'chip': {
                    'qubits': {},
                    'couplers': {},
                    'readoutLines': {},
                    'params': {},
                    'status': {},
                    'calibrations': {
                        'fluxCrosstalk': [],
                        'rfCrosstalk': []
                    }
                },
                'gates': {
                    'Measure': {},
                    'Reset': {},
                    'rfUnitary': {},
                    'CZ': {},
                    'iSWAP': {},
                    'CR': {
                        '__order_senstive__': True
                    }
                }
            })

    def newObject(self,
                  section,
                  key,
                  template=None,
                  properties=('params', 'status', 'calibrations')):
        try:
            d = self.query('.'.join([section, key]))
            return d
        except:
            pass
        d = self
        for k in chain(section.split('.'), key.split('.')):
            if k not in d:
                d[k] = {}
            d = d[k]
        if template is not None:
            d.update(template)
        for k in properties:
            if k not in d:
                d[k] = {}
        return d

    def newInstrument(self, name, driver, address, **params):
        d = self.newObject('station.instruments', name, {
            'dirver': driver,
            'address': address,
            'params': params
        })
        return d

    def newChannel(self, name, instrument='', channel=1, **params):
        try:
            self.getInstrument(instrument)
        except:
            warnings.warn(f'instrument {instrument} not exist.', )
        d = self.newObject(
            'station.channels', name, {
                'port': {
                    'instrument': instrument,
                    'channel': channel,
                },
                'params': params
            })
        return d

    def newQubit(self, name, couplers=[], readoutLine='', **params):
        for coupler in couplers:
            try:
                c = self.getCoupler(coupler)
            except:
                c = self.newCoupler(coupler, [])
            c['qubits'].append(name)
        try:
            rl = self.getReadoutLine(readoutLine)
        except:
            rl = self.newReadoutLine(readoutLine, [])
        rl['qubits'].append(name)

        d = self.newObject(
            'chip.qubits', name, {
                'params': params,
                'couplers': couplers,
                'readoutLine': readoutLine,
                'channels': {}
            })
        return d

    def newCoupler(self, name, qubits, **params):
        for qubit in qubits:
            try:
                q = self.getQubit(qubit)
            except:
                q = self.newQubit(qubit)
            q['couplers'].append(name)
        d = self.newObject('chip.couplers', name, {
            'params': params,
            'qubits': qubits,
            'channels': {}
        })
        return d

    def newReadoutLine(self, name, qubits, **params):
        for qubit in qubits:
            try:
                q = self.getQubit(qubit)
            except:
                q = self.newQubit(qubit)
            q['readoutLine'] = name
        d = self.newObject('chip.readoutLines', name, {
            'params': params,
            'qubits': qubits,
            'channels': {}
        })
        return d

    def newGate(self, name, *qubits, type='default', **params):
        for qubit in qubits:
            try:
                self.getQubit(qubit)
            except:
                self.newQubit(qubit)
        key = ','.join(qubits)
        template = {'type': type, 'params': params}
        g = self.newObject(f'gates.{name}',
                           key,
                           template,
                           properties=('type', 'params'))
        return g

    def getObject(self, q, cls=None):
        if cls is None:
            cls = 'ConfigObject'
        return self.query(q + ':' + cls)

    def getQubit(self, name):
        return self.getObject(f"chip.qubits.{name}", cls='Qubit')

    def getCoupler(self, name):
        return self.getObject(f"chip.couplers.{name}")

    def getReadoutLine(self, name):
        return self.getObject(f"chip.readoutLines.{name}")

    def getChannel(self, name):
        return self.getObject(f"station.channels.{name}")

    def getInstrument(self, name):
        return self.getObject(f"station.instruments.{name}")

    def getGate(self, name, *qubits):
        if name not in self['gates']:
            raise KeyError(f'"{name}" gate not defined.')
        if name == 'rfUnitary':
            ret = self.getObject(f"gates.{name}.{','.join(qubits)}",
                                 cls='rfUnitary')
            ret['qubits'] = tuple(qubits)
            return ret
        elif name == 'Measure':
            ret = self.getObject(f"gates.{name}.{','.join(qubits)}",
                                 cls='Measure')
            ret['qubits'] = tuple(qubits)
            return ret
        if ('__order_senstive__' in self['gates'][name]
                and self['gates'][name]['__order_senstive__']):
            ret = self.getObject(f"gates.{name}.{','.join(qubits)}",
                                 cls='Gate')
            ret['qubits'] = tuple(qubits)
            return ret
        else:
            for qlist in permutations(qubits):
                try:
                    ret = self.getObject(f"gates.{name}.{','.join(qlist)}",
                                         cls='Gate')
                    ret['qubits'] = tuple(qlist)
                    return ret
                except:
                    pass
            else:
                raise KeyError(f'Could not find "{name}" gate for {qubits}')


class ConfigObject(Trait):
    def setParams(self, **params):
        self['params'].update(params)

    def setStatus(self, **status):
        self['status'].update(status)

    def _channelDetails(self):
        if 'channels' not in self:
            return {}

        def _getChannels(dct):
            ret = {}
            for c, v in dct.items():
                if isinstance(v, str):
                    try:
                        ret[c] = self._cfg_.getChannel(v)
                    except:
                        ret[c] = v
                elif isinstance(v, dict):
                    ret[c] = _getChannels(v)
                else:
                    ret[c] = v
            return ret

        return _getChannels(self.channels)

    def _getSectionDetails(self, section, skip):
        if section not in self:
            return None

        getMethod = {
            'qubits': self._cfg_.getQubit,
            'couplers': self._cfg_.getCoupler,
            'readoutLine': self._cfg_.getReadoutLine
        }[section]

        if isinstance(self[section], str):
            return getMethod(self[section]).details(skip + (section, ))
        else:
            return {
                k: getMethod(k).details(skip + (section, ))
                for k in self[section]
            }

    def details(self, skip=('channels', 'qubits', 'couplers', 'readoutLine')):
        ret = {}
        ret.update(self)
        if 'channels' not in skip:
            ret['channels'] = self._channelDetails()
        if 'qubits' not in skip:
            ret['qubits'] = self._getSectionDetails('qubits', skip)
        if 'couplers' not in skip:
            ret['couplers'] = self._getSectionDetails('couplers', skip)
        if 'readoutLine' not in skip:
            ret['readoutLine'] = self._getSectionDetails('readoutLine', skip)
        return BaseConfig.fromdict(ret)


class Qubit(ConfigObject):
    def details(self):
        return super().details(skip=('qubits', ))


class Gate(ConfigObject):
    def details(self):
        ret = {}
        ret.update(self)
        ret['qubits'] = {
            q: self._cfg_.getQubit(q).details()
            for q in self.qubits
        }
        return BaseConfig.fromdict(ret)

    @property
    def name(self):
        return self._key_.split('.')[1]

    @property
    def qubits(self):
        return self._key_.split('.')[-1].split(',')

    def reset(self):
        qubit = self._cfg_.getQubit(self.qubits[0])
        if self.name == 'rfUnitary':
            self.setParams(frequency=qubit.params()['f01'])
        elif self.name == 'measure':
            self.setParams(frequency=qubit.params()['fr'])
        elif self.name == 'iSWAP':
            qubit2 = self._cfg_.getQubit(self.qubits[1])
            self.setParams(frequency=abs(qubit.params()['f01'] -
                                         qubit2.params()['f01']))


class rfUnitary(Gate):
    def setAmp(self, theta, amp):
        t = theta / np.pi
        dct = {k: v for k, v in zip(*self['params']['amp'])}
        dct[t] = amp
        self['params']['amp'] = [
            sorted(dct.keys()), [dct[k] for k in sorted(dct.keys())]
        ]

    def setDuration(self, theta, duration):
        t = theta / np.pi
        dct = {k: v for k, v in zip(*self['params']['duration'])}
        dct[t] = duration
        self['params']['duration'] = [
            sorted(dct.keys()), [dct[k] for k in sorted(dct.keys())]
        ]

    def setPhase(self, phi, phase):
        p = phi / np.pi
        dct = {k: v for k, v in zip(*self['params']['phase'])}
        dct[p] = phase / np.pi
        self['params']['phase'] = [
            sorted(dct.keys()), [dct[k] for k in sorted(dct.keys())]
        ]

    def amp(self, theta):
        return np.interp(theta / np.pi, *self['params']['amp'])

    def duration(self, theta):
        return np.interp(theta / np.pi, *self['params']['duration'])

    def phase(self, phi):
        return np.pi * np.interp(phi / np.pi, *self['params']['phase'])

    def reset(self):
        qubit = self._cfg_.getQubit(self.qubits[0])
        self.setParams(frequency=qubit.params.f01)
        self['params']['amp'] = [[0, 1], [self.amp(0), self.amp(np.pi)]]
        self['params']['duration'] = [[0, 1],
                                      [self.duration(0),
                                       self.duration(np.pi)]]
        self['params']['phase'] = [[-1, 1], [-1, 1]]

    def shape(self, theta, phi=0):
        return {
            'shape': self['params'].get('shape', 'Gaussian'),
            'amp': self.amp(theta),
            'duration': self.duration(theta),
            'phase': self.phase(phi),
            'frequency': self['params']['frequency'],
            'DRAGScaling': self['params'].get('DRAGScaling', 0)
        }


class Measure(Gate):
    @property
    def wfile(self) -> Path:
        return self._cfg_._path_.parent / 'Measure' / (self.qubits[0] + '.pic')

    def setW(self, w):
        self.wfile.parent.mkdir(parents=True, exist_ok=True)
        with open(self.wfile, 'wb') as f:
            pickle.dump(w, f)

    def W(self):
        with open(self.wfile, 'rb') as f:
            return pickle.load(f)


class ConfigProxy(ABC):
    @abstractmethod
    def getQubit(self, name):
        pass

    @abstractmethod
    def getCoupler(self, name):
        pass

    @abstractmethod
    def getReadoutLine(self, name):
        pass

    @abstractmethod
    def getGate(self, name, *qubits):
        pass


def setConfig(path: Union[Path, str, ABCCompileConfigMixin]) -> None:
    factory = lambda path=Path(path): Config(path)
    set_config_factory(factory)


__all__ = ['Config', 'getConfig', 'setConfig']



import click

class App():
    def __init__(self, **kwds):
        self.kwds = kwds

    def run(self):
        click.echo("Running waveforms server")
        click.echo("Server is running")
        for key, values in self.kwds.items():
            click.echo(f"    {key} : ", nl=False)
            click.secho(f"{values}", fg='blue')
        click.pause()
        MARKER = '# Everything below is ignored\n'
        message = click.edit('\n\n' + MARKER)
        if message is not None:
            click.echo(message.split(MARKER, 1)[0].rstrip('\n'))


def create_app(**kwds):
    return App(**kwds)



from .demodulate import getFTMatrix
from .distortion import (correct_reflection, distort, exp_decay_filter,
                         extractKernel, phase_curve, predistort, reflection,
                         reflection_filter, shift, zDistortKernel)
from .func import *



from __future__ import annotations

from functools import wraps
from inspect import signature
from typing import Callable, Iterable, NamedTuple, Optional, Union

import dill

from .base import Context

_NODEFAULT = object()


class Parameter(NamedTuple):
    name: str
    type: type
    default: Union[int, float, str, None] = _NODEFAULT
    unit: str = ''
    doc: str = ''


def gate(qnum: int = 1, name: Optional[str] = None, scope: dict = None):

    def decorator(func: Callable[..., Iterable], name: str = name):
        if name is None:
            name = func.__name__
        sig = signature(func)
        anum = len(sig.parameters) - 1
        if 'scope' in sig.parameters:
            anum -= 1

        @wraps(func)
        def wrapper(qubits: Union[int, tuple[int, ...]],
                    *args,
                    scope: dict = scope,
                    **kwds):
            if isinstance(qubits, (int, str)):
                if qnum != 1:
                    raise TypeError(
                        f"gate {name} should be {qnum}-qubit gate, but 1 were given."
                    )
            elif len(qubits) != qnum:
                raise TypeError(
                    f"gate {name} should be {qnum}-qubit gate, but {len(qubits)} were given."
                )
            if len(args) + len(kwds) != anum:
                raise TypeError(
                    f"gate {name} should be {anum}-arguments gate, but {len(args)+len(kwds)} were given."
                )
            if 'scope' in sig.parameters:
                kwds['scope'] = scope
            return func(qubits, *args, **kwds)

        scope[name] = wrapper
        return wrapper

    return decorator


def opaque(name: str,
           type: str = 'default',
           params: Optional[list] = None,
           scope: dict[str, dict[str, Callable]] = None):

    params = [p if isinstance(p, Parameter) else Parameter(*p) for p in params]

    def decorator(func: Callable[..., None], name: str = name):
        sig = signature(func)

        @wraps(func)
        def wrapper(ctx: Context, qubits, *args, **kwds):
            if 'scope' in sig.parameters:
                kwds['scope'] = scope
            return func(ctx, qubits, *args, **kwds)

        if name not in scope:
            scope[name] = {}
        scope[name][type] = wrapper, params
        return wrapper

    return decorator


class Library():

    def __init__(self):
        self.parents: tuple[Library, ...] = ()
        self.gates = {}
        self.opaques = {}
        self.qasmLib = {}

    def gate(self, qnum: int = 1, name: Optional[str] = None):
        return gate(qnum=qnum, name=name, scope=self.gates)

    def opaque(self,
               name: str,
               type: str = 'default',
               params: Optional[dict] = None):
        if params is None:
            params = {}
        return opaque(name, type=type, params=params, scope=self.opaques)

    def getGate(self, name: str):
        gate = self.gates.get(name, None)
        if gate is None and len(self.parents) > 0:
            for lib in self.parents:
                gate = lib.getGate(name)
                if name is not None:
                    break
        return gate

    def getOpaque(self, name: str, type: str = 'default'):
        if name in self.opaques:
            opaque, params = self.opaques[name].get(type, (None, {}))
        else:
            opaque, params = None, {}
        if opaque is None and len(self.parents) > 0:
            for lib in self.parents:
                opaque, params = lib.getOpaque(name, type)
                if opaque is not None:
                    break
        return opaque, params

    def getQasmLib(self, name: str):
        incfile = self.qasmLib.get(name, None)
        if incfile is None and len(self.parents) > 0:
            for lib in self.parents:
                incfile = lib.getQasmLib(name)
                if name is not None:
                    break
        return incfile

    def __getstate__(self):
        state = self.__dict__.copy()
        state['gates'] = dill.dumps(state['gates'])
        state['opaques'] = dill.dumps(state['opaques'])
        return state

    def __setstate__(self, state):
        state['gates'] = dill.loads(state['gates'])
        state['opaques'] = dill.loads(state['opaques'])
        self.__dict__ = state


def libraries(*libs: Library) -> Library:
    ret = Library()
    ret.parents = libs
    return ret



import numpy as np
from scipy.optimize import curve_fit


def transmon_spectrum(x, EJ, Ec, d, offset, period):
    from waveforms import Transmon

    x = (x - offset) / period
    q = Transmon(EJ=EJ, Ec=Ec, d=d)
    if isinstance(x, (int, float, complex)):
        return q.levels(flux=x)[1] - q.levels(flux=x)[0]
    else:
        y = []
        for b in x:
            y.append(q.levels(flux=b)[1] - q.levels(flux=b)[0])
        return np.asarray(y)


def transmon_spectrum_fast(x, EJ, Ec, d, offset, period):
    from scipy.special import mathieu_a, mathieu_b
    from waveforms.quantum.transmon import Transmon

    x = (x - offset) / period
    q = 0.5 * Transmon._flux_to_EJ(x, EJ, d) / Ec

    # if ng == 0:
    #     return Ec * (mathieu_b(2, -q) - mathieu_a(0, -q))
    # if ng == 0.5:
    #     return Ec * (mathieu_b(1, -q) - mathieu_a(0, -q))
    return Ec * (mathieu_b(1, -q) + mathieu_b(2, -q) -
                 2 * mathieu_a(0, -q)) / 2


def fit_transmon_spectrum(bias,
                          f01,
                          offset=0,
                          period=1,
                          f01_max=None,
                          f01_min=None,
                          alpha=None):
    from waveforms import Transmon

    x = (bias - offset) / period

    f01_max = np.max(f01) if f01_max is None else f01_max
    f01_min = np.min(f01) if f01_min is None else f01_min
    alpha = -0.24 if alpha is None else alpha

    q = Transmon(f01_max=f01_max, f01_min=f01_min, alpha=alpha)
    EJ, Ec, d = q.EJ, q.Ec, q.d

    return curve_fit(transmon_spectrum_fast,
                     bias,
                     f01,
                     p0=[EJ, Ec, d, offset, period])


def cavity_spectrum(x, offset, period, fc, f01, g=0.05, d=0.1, Ec=0.24):
    EJS = (f01 + Ec)**2 / 8 / Ec
    F = np.pi * (x - offset) / period

    EJ = EJS * np.sqrt(np.cos(F)**2 + d**2 * np.sin(F)**2)
    Delta = np.sqrt(8 * EJ * Ec) - Ec - fc
    a, b = Delta - np.sqrt(Delta**2 + g**2), Delta + np.sqrt(Delta**2 + g**2)
    ret = np.zeros_like(a)
    mask = np.abs(a) > np.abs(b)
    ret[mask] = b[mask]
    mask = np.abs(a) <= np.abs(b)
    ret[mask] = a[mask]
    return ret + fc



from numpy import inf, pi
from waveforms import Waveform, sin, square
from waveforms.qlisp.parse import *


def test_lisp_eval():
    stack = Stack()
    env = standard_env()

    eval(parse("""
        (define double (lambda (x) (+ x x)))
    """), env, stack)
    assert stack.pop() is None

    eval(
        parse("""
        (defun f (x y)
            (setq x (+ x y))
            (* x y))
    """), env, stack)
    assert stack.pop() is None

    eval(parse("""
        (double 2)
    """), env, stack)
    assert stack.pop() == 4

    eval(parse("""
        (f 2 6)
    """), env, stack)
    assert stack.pop() == 48

    eval(
        parse("""
        (let
        ((x 1)
        (y 2))
        (while (< x 10)       ; while loop
            (begin            ; test comment
            (setq x (+ x 1))
            (setq y (+ x y))
            (display! x y (f x y))
            (* x y))))
    """), env, stack)
    assert stack.pop() == 560

    eval(parse("""
        'double
    """), env, stack)
    assert stack.pop() == Symbol('double')

    eval(parse("""
        ((lambda (x) (* x x)) 11)
    """), env, stack)
    assert stack.pop() == 121

    def f(x):
        yield ('!set_waveform', Channel(('Q1', ), 'test'), sin(12e6))
        yield ('!set_phase', ('quote', 'Q1'), ('+', x, x))
        return ('double', x)

    eval(Expression((f, 8)), env, stack)
    assert stack.pop() == 16
    assert Channel(('Q1', ), 'test') in stack.raw_waveforms
    assert stack.raw_waveforms[Channel(('Q1', ), 'test')] == sin(12e6)
    assert stack.phases_ext[Symbol('Q1')][1] == 16

    eval(
        parse("""
        (waveform inf -inf nil nil nil
            3 -1e-07 0 5e-07 1 1.0 2 1 3 2 1.20112240878645e-07
            2e-07 1 3 4 100000000.0 0.0 inf 0)
    """), env, stack)
    assert stack.pop() == Waveform.fromlist([
        inf, -inf, None, None, None, 3, -1e-07, 0, 5e-07, 1, 1.0, 2, 1, 3, 2,
        1.20112240878645e-07, 2e-07, 1, 3, 4, 100000000.0, 0.0, inf, 0
    ])

    eval(
        parse("""
        (* (waveform "square(10)")
        (waveform "sin(10)"))
    """), env, stack)
    assert stack.pop() == (square(10) * sin(10))


def test_gate():
    from waveforms import wave_eval
    stack = Stack()
    env = standard_env()

    eval(
        parse("""
        (gate Y (qubit) ((rfUnitary pi (/ pi 2)) qubit))
    """), env, stack)
    assert stack.pop() is None

    eval(
        parse("""
        (gate (rfUnitary theta phi) (qubit)
            (!set_waveform (channel qubit "RF") (waveform "gaussian(20e-9) * sin(4e9)"))
            (!set_phase (channel qubit "RF") phi)
            )
    """), env, stack)
    assert stack.pop() is None

    eval(parse("""
        (rfUnitary pi 0)
    """), env, stack)
    gate = stack.pop()
    assert gate.name == 'rfUnitary'
    assert gate.params == ['theta', 'phi']
    assert gate.bindings == {'theta': pi, 'phi': 0}

    eval(parse("""
        (Y 'Q1)
    """), env, stack)
    assert stack.pop() is None
    assert stack.raw_waveforms[Channel(
        qubits='RF',
        name=('qubit', ))] == wave_eval("gaussian(20e-9) * sin(4e9)")
    assert stack.phases_ext[Channel(qubits='RF',
                                    name=('qubit', ))][1] == pi / 2



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Node for an OPENQASM opaque gate declaration."""

import warnings

from .node import Node


class Opaque(Node):
    """Node for an OPENQASM opaque gate declaration.
    children[0] is an id node.
    If len(children) is 3, children[1] is an expressionlist node,
    and children[2] is an idlist node.
    Otherwise, children[1] is an idlist node.
    """

    def __init__(self, children):
        """Create the opaque gate node."""
        super().__init__('opaque', children, None)
        self.id = children[0]  # pylint: disable=invalid-name
        # The next three fields are required by the symbtab
        self.name = self.id.name
        self.line = self.id.line
        self.file = self.id.file
        if len(children) == 3:
            self.arguments = children[1]
            self.bitlist = children[2]
        else:
            self.arguments = None
            self.bitlist = children[1]

    def n_args(self):
        """Return the number of parameter expressions."""
        if self.arguments:
            return self.arguments.size()
        return 0

    def n_bits(self):
        """Return the number of qubit arguments."""
        return self.bitlist.size()

    def qasm(self, prec=None):
        """Return the corresponding OPENQASM string."""
        if prec is not None:
            warnings.warn('Parameter \'Opaque.qasm(..., prec)\' is no longer used and is being '
                          'deprecated.', DeprecationWarning, 2)
        string = "opaque %s" % self.name
        if self.arguments is not None:
            string += "(" + self.arguments.qasm() + ")"
        string += " " + self.bitlist.qasm() + ";"
        return string


import json
import pickle
from abc import ABC, abstractmethod
from threading import Lock
from typing import Any

import portalocker

from .dicttree import NOTSET, query_tree, update_tree


class NamespaceDriver(ABC):
    @abstractmethod
    def query(self, key: str) -> Any:
        pass

    @abstractmethod
    def keys(self, pattern: str = '*') -> list[str]:
        pass

    @abstractmethod
    def update_many(self, key_value_pairs: dict[str, Any]):
        pass

    @abstractmethod
    def create_many(self, key_value_pairs: dict[str, Any]):
        pass

    @abstractmethod
    def delete_many(self, keys: list[str]):
        pass

    def close(self):
        pass

    def commit(self):
        pass


class DictDriver(NamespaceDriver):
    def __init__(self, dct):
        self.dct = dct

    def query(self, key: str) -> Any:
        return query_tree(key, self.dct)

    def keys(self, pattern: str = '*') -> list[str]:
        if not pattern.endswith('*'):
            pattern += '.*'
        keys = pattern.split('.')
        if keys[0].endswith('*'):
            d = self.dct
        else:
            d = query_tree('.'.join(keys[:-1]), self.dct)
        if not isinstance(d, dict):
            return []
        ret = d.keys()
        if keys[-1] == '*':
            return ret
        else:
            s = keys[-1].removesuffix('*')
            return [k for k in ret if k.startswith(s)]

    def update_many(self, key_value_pairs: dict[str, Any]):
        for key, value in key_value_pairs.items():
            keys = list(reversed(key.split('.')))
            for k in keys[:-1]:
                value = {k: value}
            update_tree(self.dct, {keys[-1]: value})

    def create_many(self, key_value_pairs: dict[str, Any]):
        self.update_many(key_value_pairs)

    def delete_many(self, keys: list[str]):
        for key in keys:
            ks = key.split('.')
            d = self.dct
            for k in ks[:-1]:
                if k in d:
                    d = d[k]
                else:
                    break
            else:
                try:
                    del d[ks[-1]]
                except KeyError:
                    pass


class JSONDriver(DictDriver):
    def __init__(self, path):
        super().__init__(dict())
        self.path = path
        self.load()

    def load(self):
        try:
            with open(self.path, 'r') as f:
                self.dct = json.load(f)
        except:
            self.dct = {}

    def save(self):
        with open(self.path, 'w') as f:
            json.dump(self.dct, f)

    def commit(self):
        self.save()


class PickleDriver(DictDriver):
    def __init__(self, path):
        super().__init__(dict())
        self.path = path
        self.load()

    def load(self):
        try:
            with portalocker.Lock(self.path, 'rb') as f:
                self.dct = pickle.load(f)
        except:
            self.dct = {}

    def save(self):
        with portalocker.Lock(self.path, 'wb') as f:
            pickle.dump(self.dct, f)

    def commit(self):
        self.save()


class NamespaceCache():
    def __init__(self):
        self.caches = {}
        self.notset_caches = set()
        self.fully_cached_keys = set()
        self.keys_caches = {}
        self.lock = Lock()

    def cache(self, key, value):
        with self.lock:
            self._cache(key, value)

    def delete(self, key):
        with self.lock:
            self._delete(key)

    def query(self, key):
        with self.lock:
            return self._query(key)

    def keys(self, pattern='*'):
        with self.lock:
            return self._keys(pattern)

    def clear(self):
        with self.lock:
            self._clear()

    def _cache(self, key: str, value: Any):
        if isinstance(value, tuple) and value[0] is NOTSET:
            self._cache_notset(value[1])
            return
        self._cache_keys(key, value)
        if key == '.':
            self.caches = value
            return
        keys = key.split('.')
        for k in reversed(keys):
            value = {k: value}
        update_tree(self.caches, value)

    def _fully_cached(self, key: str) -> bool:
        if key in self.fully_cached_keys:
            return True
        for k in self.notset_caches:
            if key.startswith(k + '.') or key == k:
                return True
        return False

    def _cache_notset(self, key: str):
        self.notset_caches.add(key)

    def _cache_keys(self, key: str, value: Any):
        s = set()
        for k in self.notset_caches:
            if not key.startswith(k + '.') and key != k:
                s.add(k)
        self.notset_caches = s
        self.fully_cached_keys.add(key)
        if isinstance(value, dict):
            for k, v in value.items():
                if key == '.':
                    self._cache_keys(k, v)
                else:
                    self._cache_keys(key + '.' + k, v)

    def _delete(self, key: str):
        if key == '.':
            self._clear()
            return
        keys = key.split('.')
        parent = self.caches
        d = self.caches
        for k in keys:
            try:
                parent = d
                d = d[k]
            except KeyError:
                return
        del parent[keys[-1]]
        self._delete_keys(key)

    def _delete_keys(self, key: str):
        self.notset_caches.add(key)
        s = set()
        for k in self.fully_cached_keys:
            if not k.startswith(key + '.') and k != key:
                s.add(k)
        self.fully_cached_keys = s

    def _query(self, key: str) -> tuple[Any, bool]:
        for k in self.notset_caches:
            if key.startswith(k + '.') or key == k:
                return (NOTSET, k), True
        ret = query_tree(key, self.caches)
        if isinstance(ret, tuple) and ret[0] is NOTSET:
            return None, False
        return ret, self._fully_cached(key)

    def _keys(self, pattern='*'):
        if not pattern.endswith('*'):
            pattern += '.*'
        keys = pattern.split('.')
        d, fully_cached = self._query('.'.join(keys[:-1]))
        if not fully_cached:
            return None
        if not isinstance(d, dict):
            return []
        ret = d.keys()
        if keys[-1] == '*':
            return ret
        else:
            s = keys[-1].removesuffix('*')
            return [k for k in ret if k.startswith(s)]

    def _clear(self):
        self.caches.clear()
        self.fully_cached_keys.clear()
        self.keys_caches.clear()
        self.notset_caches.clear()

    def __contains__(self, key: str):
        return key in self.fully_cached_keys


class NamespaceEventLog():
    def __init__(self):
        self.updates = NamespaceCache()
        self.creates = NamespaceCache()
        self.deletes = NamespaceCache()



import numpy as np

EPS = 1e-17


def _cmp(x, goal=0, eps=EPS):
    return np.where(x - goal > eps, 1, 0) + np.where(goal - x > eps, -1, 0)


def _point_det(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    return a.real * b.imag - a.imag * b.real


def point_on_segment(a: np.ndarray,
                     s: np.ndarray,
                     t: np.ndarray,
                     eps: float = EPS) -> np.ndarray:
    """
    a: point array, 1d of complex
    b, c: one point and another point of a segment
    """

    return (_cmp(_point_det(a - s, t - s), 0, eps)
            == 0) * (_cmp(_point_det(a - s, a - t), 0, eps) <= 0)


def point_in_polygon(a: np.ndarray,
                     b: np.ndarray,
                     eps: float = EPS) -> np.ndarray:
    """
    a: point array, 1d of complex
    b: vertex of a polygon, 1d of complex
    """

    ans = np.zeros_like(a)
    N = b.shape[0]

    for i in range(N):
        ans += point_on_segment(a, b[i], b[(i + 1) % N], eps)
        k = _cmp(_point_det(b[(i + 1) % N] - b[i], a - b[i]), 0, eps)
        d1 = _cmp((b[i] - a).imag, 0, eps)
        d2 = _cmp((b[(i + 1) % N] - a).imag, 0, eps)
        ans += np.where((k > 0) * (d1 <= 0) * (d2 > 0), 1, 0) + np.where(
            (k < 0) * (d2 <= 0) * (d1 > 0), -1, 0)

    return ans > 0



from waveforms.qlisp.simulator import (apply_circuit, apply_gate, applySeq,
                                       circuit2mat, circuit_network, gate2mat,
                                       place_at, reduceSubspace,
                                       regesterGateMatrix, seq2mat, splite_at)
from waveforms.qlisp.simulator.mat import U, fSim, make_immutable, rfUnitary



"""Multiple architecture support"""
from typing import Callable, NamedTuple

from waveforms.dicttree import flattenDict

from ..base import QLispCode
from ..commands import CommandList, DataMap, RawData, Result


class Architecture(NamedTuple):
    name: str
    description: str
    assembly_code: Callable[[QLispCode], tuple[CommandList, DataMap]]
    assembly_data: Callable[[RawData, DataMap], Result]


general_architecture = Architecture(
    name='general',
    description='General architecture',
    assembly_code=lambda code: (
        [],
        {
            'arch': 'general'
        },
    ),
    assembly_data=lambda data, _: flattenDict(data),
)

__regested_architectures = {}


def get_arch(name: str = 'general') -> Architecture:
    return __regested_architectures[name]


def register_arch(arch: Architecture):
    __regested_architectures[arch.name] = arch


register_arch(general_architecture)

__all__ = ['Architecture', 'get_arch', 'register_arch']



import itertools
import random
from typing import Callable, Iterable, Optional, Union

import numpy as np
from numpy import pi
from waveforms.quantum.rb import replace_qubit


def uncorrelatedEntropy(D: int) -> float:
    from scipy.special import polygamma

    return np.euler_gamma + polygamma(0, D)


def PTEntropy(N: int) -> float:
    D = 2**N
    return np.sum([1 / i for i in range(1, D + 1)]) - 1


def crossEntropy(P: np.array,
                 Q: np.array,
                 func: Callable = np.log,
                 eps: float = 1e-9) -> float:
    mask = (P > 0) * (Q > eps)
    if isinstance(P, (int, float, complex)):
        return -np.real(P) * np.sum(func(Q[mask]))
    else:
        return -np.sum(P[mask] * func(Q[mask]))


def Fxeb(Pm_lst: Iterable[np.array],
         Pe_lst: Iterable[np.array],
         Pi: Optional[np.array] = None) -> float:
    """
    XEB Fidelity

    Pm_lst: list of measured distribution
    Pe_lst: list of expected distribution
    """
    Si, Sm, Se = [], [], []
    for Pm, Pe in zip(Pm_lst, Pe_lst):
        if Pi is None:
            Pi = 1.0 / Pm.size
        Si.append(crossEntropy(Pi, Pe))
        Sm.append(crossEntropy(Pm, Pe))
        Se.append(crossEntropy(Pe, Pe))
    return (np.mean(Si) - np.mean(Sm)) / (np.mean(Si) - np.mean(Se))


def specklePurity(Pm_lst: Iterable[np.array]) -> float:
    D = Pm_lst[0].size
    return np.asarray(Pm_lst).var() * D**2 * (D + 1) / (D - 1)


def generateXEBCircuit(qubits: Union[int, str, tuple],
                       cycle: int,
                       seed: Optional[int] = None,
                       interleaves: list[list[tuple]] = [],
                       base: list[Union[str, tuple]] = [
                           ('rfUnitary', pi / 2, 0),
                           ('rfUnitary', pi / 2, pi / 4),
                           ('rfUnitary', pi / 2, pi / 2),
                           ('rfUnitary', pi / 2, pi * 3 / 4),
                           ('rfUnitary', pi / 2, pi),
                           ('rfUnitary', pi / 2, pi * 5 / 4),
                           ('rfUnitary', pi / 2, pi * 3 / 2),
                           ('rfUnitary', pi / 2, pi * 7 / 2)
                       ]):
    """Generate a random XEB circuit.

    Args:
        qubits (list): The qubits to use.
        cycle (int): The cycles of sequence.
        seed (int): The seed for the random number generator.
        interleaves (list): The interleaves to use.

    Returns:
        list: The XEB circuit.
    """
    if isinstance(qubits, (str, int)):
        qubits = (qubits, )

    interleaves = itertools.cycle(interleaves)

    ret = []
    rng = random.Random(seed)

    for _ in range(cycle):
        try:
            int_seq = next(interleaves)
        except StopIteration:
            int_seq = []
        ret.extend(replace_qubit(int_seq, qubits))
        for q in qubits:
            ret.append((rng.choice(base), q))

    return ret



from functools import lru_cache

import numpy as np
from scipy.linalg import eigh_tridiagonal, eigvalsh_tridiagonal
from scipy.optimize import minimize
from waveforms.math.signal import complexPeaks


class Transmon():
    def __init__(self, **kw):
        self.Ec = 0.2
        self.EJ = 20

        self.d = 0
        if kw:
            self._set_params(**kw)

    def _set_params(self, **kw):
        if {"EJ", "Ec", "d"} <= set(kw):
            return self._set_params_EJS_Ec_d(kw['EJ'], kw['Ec'], kw['d'])
        elif {"EJ", "Ec"} <= set(kw):
            return self._set_params_EJ_Ec(kw['EJ'], kw['Ec'])
        elif {"f01", "alpha"} <= set(kw):
            if 'ng' not in kw:
                return self._set_params_f01_alpha(kw['f01'], kw['alpha'])
            else:
                return self._set_params_f01_alpha(kw['f01'], kw['alpha'],
                                                  kw['ng'])
        elif {"f01_max", "f01_min"} <= set(kw):
            if {"alpha1", "alpha2"} <= set(kw):
                return self._set_params_f01_max_min_alpha(
                    kw['f01_max'], kw['f01_min'], kw['alpha1'], kw['alpha2'],
                    kw.get('ng', 0))
            elif {"alpha"} <= set(kw):
                return self._set_params_f01_max_min_alpha(
                    kw['f01_max'], kw['f01_min'], kw['alpha'], kw['alpha'],
                    kw.get('ng', 0))
            elif {"alpha1"} <= set(kw):
                return self._set_params_f01_max_min_alpha(
                    kw['f01_max'], kw['f01_min'], kw['alpha1'], kw['alpha1'],
                    kw.get('ng', 0))
        raise TypeError('_set_params() got an unexpected keyword arguments')

    def _set_params_EJ_Ec(self, EJ, Ec):
        self.Ec = Ec
        self.EJ = EJ

    def _set_params_EJS_Ec_d(self, EJS, Ec, d):
        self.Ec = Ec
        self.EJ = EJS
        self.d = d

    def _set_params_f01_alpha(self, f01, alpha, ng=0):
        Ec = -alpha
        EJ = (f01 - alpha)**2 / 8 / Ec

        def err(x, target=(f01, alpha)):
            EJ, Ec = x
            levels = self._levels(Ec, EJ, ng=ng)
            f01 = levels[1] - levels[0]
            f12 = levels[2] - levels[1]
            alpha = f12 - f01
            return (target[0] - f01)**2 + (target[1] - alpha)**2

        ret = minimize(err, x0=[EJ, Ec])
        self._set_params_EJ_Ec(*ret.x)

    def _set_params_f01_max_min_alpha(self,
                                      f01_max,
                                      f01_min,
                                      alpha1,
                                      alpha2=None,
                                      ng=0):
        if alpha2 is None:
            alpha2 = alpha1

        Ec = -alpha1
        EJS = (f01_max - alpha1)**2 / 8 / Ec
        d = (f01_min + Ec)**2 / (8 * EJS * Ec)

        def err(x, target=(f01_max, alpha1, f01_min, alpha2)):
            EJS, Ec, d = x
            levels = self._levels(Ec, self._flux_to_EJ(0, EJS, d), ng=ng)
            f01_max = levels[1] - levels[0]
            f12 = levels[2] - levels[1]
            alpha1 = f12 - f01_max

            levels = self._levels(Ec, self._flux_to_EJ(0.5, EJS, d), ng=ng)
            f01_min = levels[1] - levels[0]
            f12 = levels[2] - levels[1]
            alpha2 = f12 - f01_min

            return (target[0] - f01_max)**2 + (target[1] - alpha1)**2 + (
                target[2] - f01_min)**2 + (target[3] - alpha2)**2

        ret = minimize(err, x0=[EJS, Ec, d])
        self._set_params_EJS_Ec_d(*ret.x)

    @staticmethod
    def _flux_to_EJ(flux, EJS, d=0):
        F = np.pi * flux
        EJ = EJS * np.sqrt(np.cos(F)**2 + d**2 * np.sin(F)**2)
        return EJ

    @staticmethod
    def _levels(Ec, EJ, ng=0.0, gridSize=51, select_range=(0, 10)):
        n = np.arange(gridSize) - gridSize // 2
        w = eigvalsh_tridiagonal(4 * Ec * (n - ng)**2,
                                 -EJ / 2 * np.ones(gridSize - 1),
                                 select='i',
                                 select_range=select_range)
        return w

    @lru_cache(maxsize=128)
    def levels(self, flux=0, ng=0):
        return self._levels(self.Ec, self._flux_to_EJ(flux, self.EJ, self.d),
                            ng)

    @property
    def EJ1_EJ2(self):
        return (1 + self.d) / (1 - self.d)

    def chargeParityDiff(self, flux=0, ng=0, k=0):
        a = self.levels(flux, ng=0 + ng)
        b = self.levels(flux, ng=0.5 + ng)

        return (a[1 + k] - a[k]) - (b[1 + k] - b[k])


class FakeQPU():
    def __init__(self,
                 N,
                 EJ=15e9,
                 Ec=220e6,
                 d=0.1,
                 EJ_error=0.01,
                 Ec_error=0.01,
                 zCrosstalkSigma=0.1,
                 seed=1234):
        np.random.seed(seed)
        self.N = N
        self.M = np.eye(N) + zCrosstalkSigma * np.random.randn(N, N)
        self.bias0 = np.random.randn(N)
        self.qubits = [
            Transmon(EJ=EJ * (1 + EJ_error * np.random.randn()),
                     Ec=Ec * (1 + Ec_error * np.random.randn()),
                     d=d) for i in range(N)
        ]
        self.fr = 6.5e9 + np.arange(N) * 20e6 + 3e6 * np.random.randn(N)
        self.g = 60e6 + 5e6 * np.random.randn(N)
        self.QL = 5000 + 100 * np.random.randn(N)
        self.Qc = 6000 + 100 * np.random.randn(N)
        self.Gamma = 0.03e6 + 1e3 * np.random.randn(N)
        self.readoutBias = np.zeros(N)
        self.driveBias = np.zeros(N)
        self.driveFrequency = np.zeros(N)
        self.driveOmega = np.zeros(N)
        self.driveDuration = np.zeros(N)
        self.readoutFrequency = np.zeros(N)
        self.fluxNoise = 0.001
        self.signalNoise = 0.05
        self.phi = 0.6 * np.random.randn(N)
        self.P1 = np.zeros(N)

    def fluxList(self, bias):
        return self.M @ bias + self.bias0 + self.fluxNoise * np.random.randn(
            self.N)

    def state(self):
        return [np.random.choice([0, 1], p=[1 - p1, p1]) for p1 in self.P1]

    def S21(self, x):
        fluxList = self.fluxList(self.readoutBias)
        state = self.state()
        levels = [q.levels(flux) for q, flux in zip(self.qubits, fluxList)]
        peaks = []
        for l, s, fr, g, QL, Qc, phi in zip(levels, state, self.fr, self.g,
                                            self.QL, self.Qc, self.phi):
            f01 = l[1] - l[0]
            f12 = l[2] - l[1]
            if s == 0:
                chi = g**2 / (f01 - fr)
            else:
                chi = 2 * g**2 / (f12 - fr) - g**2 / (f01 - fr)
            fc = fr - chi
            width = fc / (2 * QL)
            amp = -QL / np.abs(Qc) * np.exp(1j * phi)
            peaks.append((fc, width, amp))
        return complexPeaks(x, peaks, 1)

    @staticmethod
    def population(Omega, Delta, Gamma, t):
        return 0.5 * Omega / np.sqrt(Omega**2 + Delta**2) * (1 - np.exp(
            -4 / 3 * Gamma * t) * np.cos(np.sqrt(Omega**2 + Delta**2) * t))

    def calcP1(self):
        for i, (bias, freq, Omega, t) in enumerate(
                zip(self.driveBias, self.driveFrequency, self.driveOmega,
                    self.driveDuration)):
            q = self.qubits[i]
            l = q.levels(bias)
            Delta = freq - l[1] + l[0]
            self.P1[i] = self.population(Omega, Delta, self.Gamma[i], t)

    def signal(self):
        s = self.S21(self.readoutFrequency)
        return s + self.signalNoise * (np.random.randn(*s.shape) +
                                       1j * np.random.randn(*s.shape))


if __name__ == "__main__":
    q = Transmon(f01=4.2, alpha=4.010 - 4.2)

    levels = q.levels()
    f01 = levels[1] - levels[0]
    f12 = levels[2] - levels[1]

    print("chargeParityDiff:")
    for k in range(4):
        diff = q.chargeParityDiff(k=k)
        print(f"  ({k},{k+1}) diff = {diff * 1e3:8.4f} MHz",
              f"(T = {1/np.abs(diff) / 2e3:.1f} us)")
    print(
        f"EJ = {q.EJ:.4f} GHz, Ec = {q.Ec*1e3:.4f} MHz, EJ/Ec={q.EJ/q.Ec:.2f}")
    print(f"f01 = {f01:.4f} GHz, alpha = {(f12-f01)*1e3:.1f} MHz")



from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Flag, auto
from typing import Any, Literal, NamedTuple, Optional, Union

from ..waveform import Waveform, zero


class Signal(Flag):
    trace = auto()
    iq = auto()
    state = auto()

    _avg_trace = auto()
    _avg_iq = auto()
    _avg_state = auto()
    _count = auto()

    trace_avg = trace | _avg_trace

    iq_avg = iq | _avg_iq

    population = state | _avg_state
    count = state | _count
    diag = state | _count | _avg_state


def gateName(st):
    if isinstance(st[0], str):
        return st[0]
    else:
        return st[0][0]


class QLispError(SyntaxError):
    pass


class MeasurementTask(NamedTuple):
    qubit: str
    cbit: int
    time: float
    signal: Signal
    params: dict
    hardware: Union[ADChannel, MultADChannel] = None
    shift: float = 0


class AWGChannel(NamedTuple):
    name: str
    sampleRate: float
    size: int = -1
    amplitude: Optional[float] = None
    offset: Optional[float] = None
    commandAddresses: tuple = ()


class MultAWGChannel(NamedTuple):
    I: Optional[AWGChannel] = None
    Q: Optional[AWGChannel] = None
    LO: Optional[str] = None
    lo_freq: float = -1
    lo_power: Optional[float] = None


class ADChannel(NamedTuple):
    name: str
    sampleRate: float = 1e9
    trigger: str = ''
    triggerDelay: float = 0
    triggerClockCycle: float = 8e-9
    commandAddresses: tuple = ()


class MultADChannel(NamedTuple):
    I: Optional[ADChannel] = None
    Q: Optional[ADChannel] = None
    IQ: Optional[ADChannel] = None
    Ref: Optional[ADChannel] = None
    LO: Optional[str] = None
    lo_freq: float = -1
    lo_power: Optional[float] = None


class GateConfig(NamedTuple):
    name: str
    qubits: tuple
    type: str = 'default'
    params: dict = {}


class ABCCompileConfigMixin(ABC):
    """
    Mixin for configs that can be used by compiler.
    """

    @abstractmethod
    def _getAWGChannel(self, name,
                       *qubits) -> Union[AWGChannel, MultAWGChannel]:
        """
        Get AWG channel by name and qubits.
        """
        pass

    @abstractmethod
    def _getADChannel(self, qubit) -> Union[ADChannel, MultADChannel]:
        """
        Get AD channel by qubit.
        """
        pass

    @abstractmethod
    def _getGateConfig(self, name, *qubits) -> GateConfig:
        """
        Return the gate config for the given qubits.

        Args:
            name: Name of the gate.
            qubits: Qubits to which the gate is applied.
        
        Returns:
            GateConfig for the given qubits.
            if the gate is not found, return None.
        """
        pass

    @abstractmethod
    def _getAllQubitLabels(self) -> list[str]:
        """
        Return all qubit labels.
        """
        pass


__config_factory = None


def set_config_factory(factory):
    global __config_factory
    __config_factory = factory


def getConfig() -> ABCCompileConfigMixin:
    if __config_factory is None:
        raise FileNotFoundError(
            'set_config_factory(factory) must be run first.')
    else:
        return __config_factory()


@dataclass
class Context():
    cfg: ABCCompileConfigMixin = field(default_factory=getConfig)
    scopes: list[dict[str, Any]] = field(default_factory=lambda: [dict()])
    qlisp: list = field(default_factory=list)
    time: dict[str,
               float] = field(default_factory=lambda: defaultdict(lambda: 0))
    addressTable: dict = field(default_factory=dict)
    waveforms: dict[str, Waveform] = field(
        default_factory=lambda: defaultdict(zero))
    raw_waveforms: dict[tuple[str, ...], Waveform] = field(
        default_factory=lambda: defaultdict(zero))
    measures: dict[int, MeasurementTask] = field(default_factory=dict)
    phases_ext: dict[str, dict[Union[int, str], float]] = field(
        default_factory=lambda: defaultdict(lambda: defaultdict(lambda: 0)))
    biases: dict[str,
                 float] = field(default_factory=lambda: defaultdict(lambda: 0))
    end: float = 0

    @property
    def channel(self):
        return self.raw_waveforms

    @property
    def phases(self):

        class D():
            __slots__ = ('ctx', )

            def __init__(self, ctx):
                self.ctx = ctx

            def __getitem__(self, qubit):
                return self.ctx.phases_ext[qubit][1] - self.ctx.phases_ext[
                    qubit][0]

            def __setitem__(self, qubit, phase):
                self.ctx.phases_ext[qubit][
                    1] = phase + self.ctx.phases_ext[qubit][0]

        return D(self)

    @property
    def params(self):
        return self.scopes[-1]

    @property
    def vars(self):
        return self.scopes[-2]

    @property
    def globals(self):
        return self.scopes[0]

    def qubit(self, q):
        return self.addressTable[q]


@dataclass
class QLispCode():
    cfg: ABCCompileConfigMixin = field(repr=False)
    qlisp: list = field(repr=True)
    waveforms: dict[str, Waveform] = field(repr=True)
    measures: dict[int, list[MeasurementTask]] = field(repr=True)
    end: float = field(default=0, repr=True)
    signal: Signal = Signal.state
    shots: int = 1024
    arch: str = 'general'
    cbit_alias: dict[int, tuple[int, int]] = field(default_factory=dict)
    sub_code_count: int = 0


def set_context_factory(factory):
    warnings.warn('set_context_factory is deprecated', DeprecationWarning, 2)


def create_context(ctx: Optional[Context] = None, **kw) -> Context:
    if ctx is None:
        return Context(**kw)
    else:
        if 'cfg' not in kw:
            kw['cfg'] = ctx.cfg
        sub_ctx = Context(**kw)
        sub_ctx.time.update(ctx.time)
        #sub_ctx.phases.update(ctx.phases)
        sub_ctx.biases.update(ctx.biases)
        for k, v in ctx.phases_ext.items():
            sub_ctx.phases_ext[k].update(v)

        return sub_ctx



from itertools import count, product

import numpy as np


def normalize(mat):
    """
    将第一个非零的矩阵元相位转成 0，以保证仅相差一个全局相位的矩阵具有相同的表示。
    优先处理对角元，再依次处理下三角阵。
    """
    assert mat.shape[0] == mat.shape[1]
    for i in range(mat.shape[0]):
        for j in range(mat.shape[0] - i):
            if np.abs(mat[j + i, j]) > 1e-3:
                mat = mat * np.abs(mat[j + i, j]) / mat[j + i, j]
                return mat
    return 0 * mat


# 相位规范化后，2 比特 Clifford 群元的矩阵元仅可能取以下值
elms = [
    0, 0, 0, 0, 1, 1j, -1, -1j, 1 / np.sqrt(2), 1j / np.sqrt(2),
    -1 / np.sqrt(2), -1j / np.sqrt(2), 0.5, 0.5j, -0.5, -0.5j
]

# 速查表，key 为模与相位组成的元组
# 对于模，0,1,2,3 分别对应 0, 1, 1/sqrt(2), 1/2
# 对于相位， 0,1,2,3 分别对应 0, pi/2, pi, 3pi/2
elms_map = {k: v for k, v in zip(product(range(4), repeat=2), elms)}


def mat2num(mat, norm=True):
    """
    将一个 2 比特 Clifford 群元对应的矩阵转换为 64 位的整数
    
    由于规范化后矩阵元只有 13 种可能取值，故每个矩阵元可用 4 位二进制整数表示，
    4 x 4 的矩阵可以用不超过 64 位的整数表示
    """

    # 仅相隔一个全局相位的操作等价，故令第一个非零的矩阵元相位为 0，保证操作与矩阵一一对应
    if norm:
        mat = normalize(mat)

    absData, phaseData = 0, 0
    for index, (i, j) in zip(count(start=0, step=2), product(range(4),
                                                             repeat=2)):
        for k, v in elms_map.items():
            if abs(v - mat[i, j]) < 1e-3:
                a, phase = k
                break
        else:
            raise ValueError(f"Element {mat[i, j]} not allowed.")
        absData |= a << index
        phaseData |= phase << index
    return absData | (phaseData << 32)


def num2mat(num):
    """
    将 64 位整数还原成矩阵
    """
    absData, phaseData = num & 0xffffffff, num >> 32
    mat = np.zeros((4, 4), dtype=complex)

    for index, (i, j) in zip(count(start=0, step=2), product(range(4),
                                                             repeat=2)):
        a, phase = (absData >> index) & 0x3, (phaseData >> index) & 0x3
        mat[i, j] = elms_map[(a, phase)]
    return mat



import random

import numpy as np

from .circuit.simulator import seq2mat
from .clifford import cliffordOrder
from .clifford.clifford import (generateTwoQubitCliffordSequence, inv,
                                mat2index, mul)
from .clifford.seq2mat import seq2qlisp

_index2seq = [[seq] for seq in generateTwoQubitCliffordSequence()]


def replace_qubit(circuit, qubits):
    ret = []
    for gate, target in circuit:
        if isinstance(target, int):
            ret.append((gate, qubits[target]))
        else:
            ret.append((gate, tuple(qubits[i] for i in target)))
    return ret


def circuit_to_index(circuit: list) -> int:
    if not circuit:
        return 0
    mat = seq2mat(circuit)
    if mat.shape[0] == 2:
        mat = np.kron(np.eye(2), mat)
    return mat2index(mat)


def index_to_circuit(index: int, qubits=(0, ), base=None, rng=None) -> list:
    if len(qubits) > 2:
        raise ValueError('Only support 1 or 2 qubits')
    if rng is None:
        rng = random.Random()
    if base is None:
        base = _index2seq
    seq = rng.choice(base[index])
    if len(qubits) == 1:
        seq = (seq[1], )
    return seq2qlisp(seq, range(len(qubits)))


def generateRBCircuit(qubits, cycle, seed=None, interleaves=[], base=None):
    """Generate a random Clifford RB circuit.

    Args:
        qubits (list): The qubits to use.
        cycle (int): The cycles of clifford sequence.
        seed (int): The seed for the random number generator.
        interleaves (list): The interleaves to use.
        base (list): The basic two-qubit Clifford sequence.

    Returns:
        list: The RB circuit.
    """
    if isinstance(qubits, (str, int)):
        qubits = (qubits, )

    MAX = cliffordOrder(len(qubits))

    interleaves_index = circuit_to_index(interleaves)

    ret = []
    index = 0
    rng = random.Random(seed)

    for _ in range(cycle):
        i = rng.randrange(MAX)
        index = mul(i, index)
        ret.extend(index_to_circuit(i, qubits, base, rng))
        index = mul(interleaves_index, index)
        ret.extend(interleaves)

    ret.extend(index_to_circuit(inv(index), qubits, base, rng))

    return replace_qubit(ret, qubits)



from waveforms.qlisp.base import (ABCCompileConfigMixin, ADChannel, AWGChannel,
                                  Context, GateConfig, MeasurementTask,
                                  MultADChannel, MultAWGChannel, QLispCode,
                                  QLispError, Signal, create_context, gateName,
                                  getConfig, set_config_factory,
                                  set_context_factory)



import click


@click.command()
@click.option('--executor',
              default='127.0.0.1',
              help='Executor address to use.')
@click.option('--port', default=8080, help='Port to run the server on.')
@click.option('--host', default='0.0.0.0', help='Host to run the server on.')
@click.option('--db-url', default=None, help='Database URL to use.')
@click.option('--data-path',
              default='waveforms/data',
              help='Path to the data directory.')
@click.option('--debug', is_flag=True, help='Run in debug mode.')
@click.option('--workers',
              default=1,
              help='Number of workers to run the server with.')
@click.option('--timeout', default=60, help='Timeout for requests.')
@click.option('--log-level',
              default='INFO',
              help='Log level to run the server with.')
@click.option('--log-file',
              default='/var/log/waveforms/server.log',
              help='Log file to run the server with.')
@click.option('--log-format',
              default='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
              help='Log format to run the server with.')
def main(executor, port, host, db_url, data_path, debug, workers, timeout,
         log_level, log_file, log_format):
    """
    Main entry point for the server.
    """
    from waveforms.server import create_app

    app = create_app(
        executor=executor,
        port=port,
        host=host,
        db_url=db_url,
        data_path=data_path,
        debug=debug,
        workers=workers,
        timeout=timeout,
        log_level=log_level,
        log_file=log_file,
        log_format=log_format,
    )
    app.run()


if __name__ == '__main__':
    main()


from .arch import (COMMAND, FREE, PUSH, READ, SYNC, TRIG, WRITE, Architecture,
                   get_arch, register_arch)
from .arch.base import CommandList, DataMap, RawData, Result
from .compiler import compile
from .library import Library, libraries
from .qlisp import QLispCode, QLispError
from .stdlib import std



import operator
from functools import lru_cache, reduce
from itertools import chain, combinations, product, repeat
from typing import Optional

import numpy as np
from scipy import linalg, optimize
from scipy.sparse import coo_matrix, csc_matrix
from scipy.sparse.linalg import inv, lsqr

from ..cache import cache
from .math import dagger, normalize, randomUnitary, unitary2v, v2unitary

__base_op = {
    'I': np.array([[1, 0], [0, 1]], dtype=complex),
    'X': np.array([[0, -1j], [-1j, 0]], dtype=complex),
    'Y': np.array([[0, -1], [1, 0]], dtype=complex),
    'X/2': np.array([[1, -1j], [-1j, 1]], dtype=complex) / np.sqrt(2),
    'Y/2': np.array([[1, -1], [1, 1]], dtype=complex) / np.sqrt(2),
    '-X/2': np.array([[1, 1j], [1j, 1]], dtype=complex) / np.sqrt(2),
    '-Y/2': np.array([[1, 1], [-1, 1]], dtype=complex) / np.sqrt(2),
    'Z': np.array([[1, 0], [0, -1]], dtype=complex),
    'S': np.array([[1, 0], [0, 1j]], dtype=complex),
    '-S': np.array([[1, 0], [0, -1j]], dtype=complex),
    'H': np.array([[1, 1], [1, -1]], dtype=complex) / np.sqrt(2),

    # non-clifford
    'T': np.array([[1, 0], [0, 1 / np.sqrt(2) + 1j / np.sqrt(2)]]),
    '-T': np.array([[1, 0], [0, 1 / np.sqrt(2) - 1j / np.sqrt(2)]]),
}

__base_op_2 = {
    # two qubits
    'CZ':
    np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]],
             dtype=complex),
    'CNOT':
    np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]],
             dtype=complex),
    'iSWAP':
    np.array([[1, 0, 0, 0], [0, 0, 1j, 0], [0, 1j, 0, 0], [0, 0, 0, 1]],
             dtype=complex),
}

__base_op['Sigma_x'] = 1j * __base_op['X']
__base_op['Sigma_y'] = 1j * __base_op['Y']
__base_op['Sigma_z'] = __base_op['Z']
__base_op['Sigma_p'] = (__base_op['Sigma_x'] - 1j * __base_op['Sigma_y']) / 2
__base_op['Sigma_m'] = (__base_op['Sigma_x'] + 1j * __base_op['Sigma_y']) / 2

qst_gates = ['-Y/2', 'X/2', 'I']
ocqst_gates = ['I', 'X/2', 'Y/2', '-X/2', '-Y/2', 'X']
qpt_init_gates = ['I', 'X', 'Y/2', 'X/2']
pauli_basis = ['I', 'Sigma_x', 'Sigma_y', 'Sigma_z']
real_pauli_basis = ['I', 'Sigma_x', 'Y', 'Z']
raise_lower_basis = ['I', 'Sigma_p', 'Sigma_m', 'Sigma_z']


def tensorMatrix(transform):
    """transform 所对应变换矩阵
    """
    return reduce(np.kron, (__base_op[k] for k in transform))


@lru_cache()
def tensorElement(opList, r, c):
    """直积矩阵第 r 行、第 c 列元素
    """
    N = len(opList)
    return reduce(
        operator.mul,
        (__base_op[k][int(x)][int(y)]
         for k, x, y in zip(opList, f'{r:b}'.zfill(N), f'{c:b}'.zfill(N))))


def rhoToV(rho):
    """密度矩阵转成一个纯实数向量
    """
    N = rho.shape[0]
    indices = np.triu_indices(N, 1)
    return np.hstack(
        (rho[indices].real, rho[indices].imag, np.diag(rho)[1:].real))


def vToRho(V):
    """rhoToV 的逆函数"""
    N = int(np.sqrt(len(V) + 1))
    assert N**2 - 1 == len(V)

    X = V[:(N**2 - N) // 2]
    Y = V[(N**2 - N) // 2:(N**2 - N)]
    Z = V[(N**2 - N):]
    rho = np.diag(np.hstack((1 - np.sum(Z), Z)) / 2).astype(complex)

    rho[np.triu_indices(N, 1)] = X + 1j * Y
    rho += dagger(rho)

    return normalize(rho)


def qstOpList(N, gates=qst_gates):
    """State tomography 对应的全部操作
    """
    return product(gates, repeat=N)


def qptInitList(N, gates=qpt_init_gates):
    return product(gates, repeat=N)


def qptBases(N, gates=pauli_basis):
    return product(gates, repeat=N)


def xFactor(i, transform, j, k):
    """经过 transform 变换后的密度矩阵第 i 个对角元表达式中 x_jk 的系数
    """
    return 2 * np.real(
        np.conj(tensorElement(transform, i, k)) *
        tensorElement(transform, i, j))


def yFactor(i, transform, j, k):
    """经过 transform 变换后的密度矩阵第 i 个对角元表达式中 y_jk 的系数
    """
    return -2 * np.imag(
        np.conj(tensorElement(transform, i, k)) *
        tensorElement(transform, i, j))


def zFactor(i, transform, j):
    """经过 transform 变换后的密度矩阵第 i 个对角元表达式中 z_j 的系数
    """
    return np.real(
        np.conj(tensorElement(transform, i, j)) *
        tensorElement(transform, i, j) -
        np.conj(tensorElement(transform, i, 0)) *
        tensorElement(transform, i, 0))


def constFactor(i, transform):
    """经过 transform 变换后的密度矩阵第 i 个对角元表达式中的常数项
    """
    return np.real(
        np.conj(tensorElement(transform, i, 0)) *
        tensorElement(transform, i, 0))


@cache()
def formUMatrix(n, gates):
    """构造从相干矢到测量结果的转移矩阵"""
    dim = 2**n
    A_data, A_i, A_j = [], [], []
    C = []

    for row, (U,
              i) in enumerate(product(qstOpList(n, gates=gates), range(1,
                                                                       dim))):
        C.append(constFactor(i, U))
        for col, (jk, func) in enumerate(
                chain(zip(combinations(range(dim), 2), repeat(xFactor)),
                      zip(combinations(range(dim), 2), repeat(yFactor)),
                      zip(combinations(range(1, dim), 1), repeat(zFactor)))):
            A = func(i, U, *jk)

            if np.abs(A) > 1e-9:
                A_i.append(row)
                A_j.append(col)
                A_data.append(A)
    A, b = coo_matrix((A_data, (A_i, A_j))), np.asarray(C)
    return A, b


def qst(diags, gates=qst_gates):
    """Convert a set of diagonal measurements into a density matrix.
    
    diags - measured probabilities (diagonal elements) after acting
            on the state with each of the unitaries from the qst
            protocol.
    gates - qst protocol
            yield unitaries from product(gates, repeat=N).
            gates should be choosed from {I, X, Y, Z, H, S, -S,
            X/2, Y/2, -X/2, -Y/2}.
    """
    diags = np.asarray(diags)
    N = len(diags[0])  # size of density matrix
    n = int(np.log2(N))  # number of qubits

    A, b = formUMatrix(n, gates=gates)

    P = np.asarray(diags)[:, 1:]
    v, *_ = lsqr(A, P.flatten() - b)
    return vToRho(v)


@cache()
def _UUds(N, gates):
    return [(tensorMatrix(ops), dagger(tensorMatrix(ops)))
            for ops in qstOpList(N, gates=gates)]


def _qst_mle(diags, UUds: list[tuple[np.ndarray, np.ndarray]],
             rho0: np.ndarray, F: np.ndarray):
    """State tomography with maximum-likelihood estimation.
    
    diags - measured probabilities (diagonal elements) after acting
            on the state with each of the unitaries from the qst
            protocol.
    UUds - list of unitary pairs (U, U^\dagger)
    rho0 - initial density matrix
    F - fidelity of the measurement
    """
    # convert the initial guess into vector
    v_guess = rhoToV(rho0)

    def log(x):
        """Safe version of log that returns -Inf when x < 0, rather than NaN.
        
        This is good for our purposes since negative probabilities are infinitely unlikely.
        """
        return np.log(x.real * (x.real > 0))

    def unlikelihood(v):  # negative of likelihood function
        rho = vToRho(v)
        pxis = np.array([F @ np.diag(U @ rho @ Ud) for U, Ud in UUds])
        pxis = pxis * (pxis.real > 0)
        terms = (diags * log(pxis + (diags == 0)) +
                 (1 - diags) * log(1 - pxis + (diags == 1)))
        return -np.sum(terms.real)

    #minimize
    #tis = optimize.fmin(unlikelihood, tis_guess)
    #tis = optimize.fmin_bfgs(unlikelihood, tis_guess)
    v = optimize.minimize(unlikelihood, v_guess, method='BFGS')['x']
    return vToRho(v)


def qst_mle(diags, gates=qst_gates, F=None, rho0=None):
    """State tomography with maximum-likelihood estimation.
    
    diags - a 2D array of measured probabilites.  The first index indicates which
            operation from the qst protocol was applied, while the second index
            tells which measurement result this was (e.g. 000, 001, etc.).
    gates - qst protocol
            yield unitaries from product(gates, repeat=N).
            gates should be choosed from {I, X, Y, Z, H, S, -S,
            X/2, Y/2, -X/2, -Y/2}.
    rho0 - an initial guess for the density matrix, e.g. from linear tomography.
    """
    diags = np.asarray(diags)
    N = len(diags[0])  # size of density matrix
    n = int(np.log2(N))  # number of qubits

    # make an initial guess using linear tomography
    if rho0 is None:
        rho0 = normalize(qst(diags, gates))

    # precompute conjugate transposes of matrices
    UUds = _UUds(n, gates)

    if F is None:
        F = np.eye(N)

    return _qst_mle(diags, UUds, rho0, F)


def chi0(rhosIn, rhosOut):
    """Calculates the pointer-basis chi-matrix.
    
    rhosIn - array of input density matrices
    rhosOut - array of output density matrices.
    
    Uses linalg.lstsq to calculate the closest fit
    when the chi-matrix is overdetermined by the data.
    """
    Din = rhosIn[0].size
    Dout = rhosOut[0].size
    n = len(rhosIn)

    rhosIn_mat = np.array(rhosIn).reshape((n, Din))
    rhosOut_mat = np.array(rhosOut).reshape((n, Dout))

    chi0, resids, rank, s = linalg.lstsq(rhosIn_mat,
                                         rhosOut_mat,
                                         overwrite_a=True,
                                         overwrite_b=True)
    return chi0


@cache()
def chi0_to_chi_mat(N, basis=pauli_basis):
    return inv(chi_to_chi0_mat(N, basis))


@cache()
def chi_to_chi0_mat(N, basis=pauli_basis):
    dim = 2**N
    A_data, A_i, A_j = [], [], []
    for row, (k, l, i, j) in enumerate(product(range(dim), repeat=4)):
        for col, (m,
                  n) in enumerate(product(qptBases(N, gates=basis), repeat=2)):
            elm = tensorElement(m, i, k) * np.conj(tensorElement(n, j, l))
            if abs(elm) > 1e-12:
                A_i.append(row)
                A_j.append(col)
                A_data.append(elm)
    return csc_matrix((A_data, (A_i, A_j)), dtype=complex)


def chi0_to_chi(chi0, basis=pauli_basis):
    dim = int(round(np.sqrt(chi0.shape[0])))
    N = int(round(np.log2(dim)))
    M = chi0_to_chi_mat(N, basis)
    chi = M @ chi0.reshape(dim**4)
    return chi.reshape((dim**2, dim**2))


def chi_to_chi0(chi, basis=pauli_basis):
    dim = int(round(np.sqrt(chi.shape[0])))
    N = int(round(np.log2(dim)))
    M = chi_to_chi0_mat(N, basis)
    chi0 = M @ chi.reshape(dim**4)
    return chi0.reshape((dim**2, dim**2))


#def qpt(a, b=None, /, *, init_gates=qpt_init_gates, basis=pauli_basis):
def qpt(a, b=None, *, init_gates=qpt_init_gates, basis=pauli_basis):
    if b is None:
        rfList = a
        rho0 = np.zeros_like(rfList[0], dtype=complex)
        rho0[0, 0] = 1
        N = int(round(np.log2(rfList[0].shape[0])))
        Us = [tensorMatrix(seq) for seq in qptInitList(N, gates=init_gates)]
        riList = [U @ rho0 @ dagger(U) for U in Us]
    else:
        riList, rfList = a, b
    chi = chi0(np.array(riList), np.array(rfList))
    return chi0_to_chi(chi, basis)


def applyOp(opList, rho=None):
    if rho is None:
        rho = reduce(np.kron,
                     repeat(np.array([[1, 0], [0, 0]]), times=len(opList)))
    U = tensorMatrix(opList)
    return U @ rho @ dagger(U)


def applyChi(chi, rho, basis=pauli_basis):
    dim = rho.shape[0]
    chi0 = chi_to_chi0(chi, basis)
    rhoF = rho.reshape(dim * dim) @ chi0
    return rhoF.reshape((dim, dim))


def chain_process(chi1, chi2, basis=pauli_basis):
    chi0 = chi_to_chi0(chi2, basis) @ chi_to_chi0(chi1, basis)
    return chi0_to_chi(chi0, basis)


def U_to_chi(U, basis=pauli_basis):
    riList = []
    rfList = []

    N = int(round(np.log2(U.shape[0])))

    for Ui in qptInitList(N):
        # 分别制备不同的初态
        ri = applyOp(Ui)
        # 做完操作后测末态的 QST
        rf = U @ ri @ dagger(U)
        riList.append(ri)
        rfList.append(rf)

    # 计算 chi 超算符
    chi = chi0(np.asarray(riList), np.asarray(rfList))
    chi = chi0_to_chi(chi, basis)
    return chi


def chi_to_U(chi, basis=pauli_basis, U0=None):
    from waveforms.quantum.clifford.mat import normalize

    def unlikelihood(v):
        U = v2unitary(v)
        chi1 = U_to_chi(U, basis)
        return -np.abs(np.trace(chi @ dagger(chi1))) / np.sqrt(
            np.trace(chi @ dagger(chi)).real *
            np.trace(chi1 @ dagger(chi1)).real)
        #return (np.abs(chi - chi1)**2).sum()

    dim = int(round(np.sqrt(chi.shape[0])))

    if U0 is None:
        v_guess = unitary2v(randomUnitary(dim))
    else:
        v_guess = unitary2v(U0)
    v = optimize.minimize(unlikelihood, v_guess, method='BFGS')['x']
    return normalize(v2unitary(v))


__all__ = ["qst", "qst_mle", "qpt", "qstOpList", "qptInitList"]



import functools
import hashlib
import pathlib
import pickle
import time
from typing import Any, Callable, Hashable, KeysView, Optional, Union

try:
    from portalocker import Lock
except:
    import warnings

    warnings.warn('portalocker not installed, cache will not be thread safe!')

    class Lock():

        def __init__(self, file, mode):
            self.file = file
            self.mode = mode
            self.fh = None

        def __enter__(self):
            self.fh = open(self.file, self.mode)
            return self.fh

        def __exit__(self, exc_type, exc_value, traceback):
            self.fh.close()


cache_dir = pathlib.Path.home() / '.waveforms' / 'cache'
cache_dir.mkdir(parents=True, exist_ok=True)

MAXVALUESIZE = 1024
Decorator = Callable[[Callable], Callable]


class Cache(dict):

    def __init__(self, name: str, path: pathlib.Path = cache_dir):
        name = name.split('.')
        self.path: pathlib.Path = path.joinpath(*name[:-1]) / (name[-1] +
                                                               '.cache.d')
        self._buffer = {}
        self._index = {}
        self._mtime = 0
        (self.path / 'values').mkdir(parents=True, exist_ok=True)

    @property
    def index(self) -> pathlib.Path:
        return self.path / 'index'

    @staticmethod
    def _hash(key: Hashable) -> str:
        return hashlib.md5(pickle.dumps(key)).hexdigest()

    def _syncIndex(self) -> None:
        if not self.index.exists() or self._mtime > self.index.stat().st_mtime:
            with Lock(self.index, 'wb') as fh:
                pickle.dump(self._index, fh)
        elif self._mtime < self.index.stat().st_mtime:
            with Lock(self.index, 'rb') as fh:
                self._index = pickle.load(fh)
        self._mtime = self.index.stat().st_mtime

    def _valuePath(self, key: Hashable) -> pathlib.Path:
        hashedKey = self._hash(key)
        return self.path / 'values' / hashedKey

    def _storeValue(self, k: Hashable, buf: bytes) -> None:
        with Lock(self._valuePath(k), 'wb') as fh:
            fh.write(buf)

    def _loadValue(self, k: Hashable) -> None:
        with Lock(self._valuePath(k), 'rb') as fh:
            buf = fh.read()
        hashstr = self._hash(buf)
        self._index[k] = self._valuePath(k).stat().st_mtime, None, hashstr
        self._buffer[k] = pickle.loads(buf)

    def __setitem__(self, k: Hashable, v: Any) -> None:
        buf = pickle.dumps(v)
        if len(buf) <= MAXVALUESIZE:
            mtime = time.time()
            self._index[k] = mtime, buf, ''
            self._mtime = mtime
            self._valuePath(k).unlink(missing_ok=True)
        else:
            hashstr = self._hash(buf)
            if k not in self._index or hashstr != self._index[k][2]:
                self._storeValue(k, buf)
                mtime = self._valuePath(k).stat().st_mtime
                self._index[k] = mtime, None, hashstr
                self._mtime = mtime
        self._syncIndex()
        self._buffer[k] = v

    def __getitem__(self, k: Hashable) -> Any:
        self._syncIndex()
        mtime, buf, hashstr = self._index[k]
        if k not in self._buffer and buf is not None:
            self._buffer[k] = pickle.loads(buf)
        elif k not in self._buffer:
            self._loadValue(k)
        return self._buffer[k]

    def __contains__(self, x: Hashable) -> bool:
        self._syncIndex()
        return x in self._index

    def keys(self) -> KeysView[Hashable]:
        self._syncIndex()
        return self._index.keys()

    def clear(self) -> None:
        self._buffer.clear()
        self._index.clear()
        self.index.unlink()
        for f in (self.path / 'values').iterdir():
            f.unlink()
        self._mtime = 0


__caches = {}


def _getCache(name: str) -> Cache:
    try:
        return __caches[name]
    except:
        __caches[name] = Cache(name)
        return __caches[name]


def cache(storage: Optional[Union[str, Cache]] = None) -> Decorator:

    def decorator(func: Callable,
                  storage: Optional[Union[str, Cache]] = storage) -> Callable:
        if storage is None:
            storage = _getCache('.'.join([func.__module__, func.__name__]))
        elif isinstance(storage, str):
            storage = _getCache(storage)
        elif not isinstance(storage, Cache):
            raise Exception(f'storage type {type(storage)} not understand!')

        @functools.wraps(func)
        def wrapper(*args, **kwds):
            kwds = {k: kwds[k] for k in sorted(kwds)}
            try:
                ret = storage[pickle.dumps((args, kwds))]
            except:
                ret = func(*args, **kwds)
                storage[pickle.dumps((args, kwds))] = ret
            return ret

        wrapper.cache = storage
        return wrapper

    return decorator


def clear(name: str) -> None:
    cache = _getCache(name)
    cache.clear()



from __future__ import annotations

import functools
import inspect
import math
import operator as op
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, NamedTuple, Union

import numpy as np
from waveforms.waveform import Waveform, zero

from .base import ABCCompileConfigMixin, ADChannel, MultADChannel
from .tokenize import Expression, Number, Symbol, tokenize

################ Parsing


def parse(program):
    "Read a Scheme expression from a string."

    stack = []
    end = False

    for token in tokenize(program):
        if end:
            raise SyntaxError('unexpected EOF while reading')
        if token.type == '(':
            stack.append([])
        elif token.type == 'QUOTE':
            stack.append([Symbol('quote')])
        elif token.type == ')':
            try:
                s_expr = Expression(stack.pop())
            except IndexError:
                raise SyntaxError(
                    f'unexpected ) at L{token.line}C{token.column}')
            while True:
                if stack:
                    stack[-1].append(s_expr)
                    if Symbol('quote') == stack[-1][0]:
                        s_expr = Expression(stack.pop())
                    else:
                        break
                else:
                    end = s_expr
                    break
        else:
            s_expr = token.value
            while True:
                if stack:
                    stack[-1].append(s_expr)
                    if Symbol('quote') == stack[-1][0]:
                        s_expr = Expression(stack.pop())
                    else:
                        break
                elif Symbol('quote') == s_expr[0]:
                    end = s_expr
                    break
                else:
                    raise SyntaxError(
                        f'expected ( before L{token.line}C{token.column}')
    if not end:
        raise SyntaxError(f'expected ) after L{token.line}C{token.column}')
    return end


################ Environments


class Env(dict):
    "An environment: a dict of {'var':val} pairs, with an outer Env."

    def __init__(self, params=(), args=(), outer=None):
        self.update(zip(params, args))
        self.outer = outer

    def find(self, var):
        "Find the innermost Env where var appears."
        return self if (var in self) else self.outer.find(var)

    def lookup(self, name):
        "Find the value of a variable, starting in this Env."
        try:
            return self.find(name)[name]
        except:
            raise KeyError(f'can not find {name}')

    def set(self, name, value):
        "Set a variable to a value."
        self.find(name)[name] = value

    def assign(self, name, value):
        "Set a variable to a value."
        self[name] = value


################ Procedures


class Procedure():
    "A user-defined Scheme procedure."

    def __init__(self, params, body, env):
        self.params, self.body, self.env = params, body, env


class Gate(Procedure):
    "A quantum operation."

    def __init__(self, name, params, qubits, body, env, bindings=None):
        self.name = name
        self.params = params
        self.bindings = bindings
        self.qubits = qubits
        self.body = body
        self.env = env


class Channel(NamedTuple):
    qubits: tuple
    name: str


################ eval


class LispError(Exception):
    pass


def error(msg, *args):
    raise LispError(msg, *args)


def make_waveform(*args):
    from waveforms import Waveform, wave_eval

    if len(args) == 1 and isinstance(args[0], str):
        return wave_eval(args[0])
    elif len(args) == 1 and isinstance(args[0], Waveform):
        return args[0]
    else:
        try:
            return Waveform.fromlist(list(args))
        except:
            raise RuntimeError('invalid waveform')


def standard_env():
    "An environment with some Scheme standard procedures."
    env = Env()
    env.update(vars(math))  # sin, cos, sqrt, pi, ...
    env.update({
        '+': lambda *x: functools.reduce(op.add, x),
        '*': lambda *x: functools.reduce(op.mul, x),
        '&': lambda *x: functools.reduce(op.and_, x),
        '|': lambda *x: functools.reduce(op.or_, x),
        '-': op.sub, '/': op.truediv, '%': op.mod, '**': op.pow,
        '>': op.gt, '<': op.lt, '>=': op.ge, '<=': op.le, '==': op.eq,
        '>>': op.rshift, '<<': op.lshift, '^': op.xor,
        'pow':     op.pow,
        'abs':     abs,
        'append':  op.add,
        'car':     lambda x: x[0],
        'cdr':     lambda x: x[1:],
        'cons':    lambda x,y: (x,) + y,
        'is?':     op.is_,
        'eq?':     op.eq,
        'equal?':  op.eq,
        'length':  len,
        'list':    lambda *x: x,
        'list?':   lambda x: isinstance(x, tuple),
        'max':     max,
        'min':     min,
        'and':     all,
        'or':      any,
        'all':     all,
        'any':     any,
        'not':     op.not_,
        'null?':   lambda x: x == (),
        'number?': lambda x: isinstance(x, Number),
        'procedure?': lambda x: isinstance(x, Procedure) or callable(x),
        'round':   round,
        'symbol?': lambda x: isinstance(x, Symbol),
        'string?': lambda x: isinstance(x, str),
        'display!':print,
        'error!':  error,
        'waveform':make_waveform,
        'pi':      np.pi,
        'e':       np.e,
        'inf':     np.inf,
        '-inf':    -np.inf,
        'nil':     None,
        'null':    None,
        'None':    None,
        'true':    True,
        'false':   False,
        'True':    True,
        'False':   False,
    }) # yapf: disable
    return env


class Capture(NamedTuple):
    qubit: str
    cbit: int
    time: float
    signal: Union[str, tuple[str]]
    params: dict
    hardware: Union[ADChannel, MultADChannel] = None
    shift: float = 0


@dataclass
class Stack():
    cfg: ABCCompileConfigMixin = field(default_factory=dict)
    scopes: list[dict[str, Any]] = field(default_factory=lambda: [dict()])
    qlisp: list = field(default_factory=list)
    time: dict[str,
               float] = field(default_factory=lambda: defaultdict(lambda: 0))
    addressTable: dict = field(default_factory=dict)
    waveforms: dict[str, Waveform] = field(
        default_factory=lambda: defaultdict(zero))
    raw_waveforms: dict[tuple[str, ...], Waveform] = field(
        default_factory=lambda: defaultdict(zero))
    measures: dict[int, Capture] = field(default_factory=dict)
    phases_ext: dict[str, dict[Union[int, str], float]] = field(
        default_factory=lambda: defaultdict(lambda: defaultdict(lambda: 0)))
    biases: dict[str,
                 float] = field(default_factory=lambda: defaultdict(lambda: 0))
    end: float = 0
    _ret: list = field(default_factory=list)

    @property
    def channel(self):
        return self.raw_waveforms

    @property
    def phases(self):

        class D():
            __slots__ = ('ctx', )

            def __init__(self, ctx):
                self.ctx = ctx

            def __getitem__(self, qubit):
                return self.ctx.phases_ext[qubit][1]

            def __setitem__(self, qubit, phase):
                self.ctx.phases_ext[qubit][1] = phase

        return D(self)

    @property
    def params(self):
        return self.scopes[-1]

    @property
    def vars(self):
        return self.scopes[-2]

    @property
    def globals(self):
        return self.scopes[0]

    def qubit(self, q):
        return self.addressTable[q]

    def push(self, x):
        self._ret.append(x)

    def pop(self):
        return self._ret.pop()

    def pick(self):
        return self._ret[-1]


def eval_quote(exp, env, stack):
    (_, exp) = exp
    stack.push(exp)


def eval_if(exp, env, stack):
    (_, test, conseq, alt) = exp
    eval(test, env, stack)
    if stack.pop():
        eval(conseq, env, stack)
    else:
        eval(alt, env, stack)


def eval_cond(exp, env, stack):
    for (test, conseq) in exp[1:]:
        eval(test, env, stack)
        if stack.pop():
            eval(conseq, env, stack)
            break
    stack.push(None)


def eval_while(exp, env, stack):
    (_, test, body) = exp
    ret = None
    while True:
        eval(test, env, stack)
        if not stack.pop():
            break
        eval(body, env, stack)
        ret = stack.pop()
    stack.push(ret)


def eval_define(exp, env, stack):
    (_, var, exp) = exp
    eval(exp, env, stack)
    if not isinstance(var, Symbol):
        raise TypeError(f'var must be a symbol')
    env.assign(var.name, stack.pop())
    stack.push(None)


def eval_set(exp, env, stack):
    (_, var, exp) = exp
    eval(exp, env, stack)
    eval(var, env, stack)
    var = stack.pop()
    if not isinstance(var, Symbol):
        raise TypeError(f'var must be a symbol')
    env.set(var.name, stack.pop())
    stack.push(None)


def eval_setq(exp, env, stack):
    (_, var, exp) = exp
    if not isinstance(var, Symbol):
        raise TypeError(f'var must be a symbol')
    eval(exp, env, stack)
    env.set(var.name, stack.pop())
    stack.push(None)


def eval_lambda(exp, env, stack):
    (_, params, body) = exp
    for i, p in enumerate(params, start=1):
        if not isinstance(p, Symbol):
            raise TypeError(f'the {i} param must be a symbol')
    stack.push(Procedure([p.name for p in params], body, env))


def eval_defun(exp, env, stack):
    (_, var, params, *body) = exp
    if not isinstance(var, Symbol):
        raise TypeError(f'var must be a symbol')
    env.assign(
        var.name,
        Procedure([p.name for p in params],
                  Expression([Symbol('begin'), *body]), env))
    stack.push(None)


def eval_gate(exp, env, stack):
    (_, var, qubits, *body) = exp
    if isinstance(var, Expression):
        gate, *params = var
    elif isinstance(var, Symbol):
        gate = var
        params = []
    else:
        raise TypeError(f'var must be a symbol')
    env.assign(
        gate.name,
        Gate(gate.name, [p.name for p in params], [q.name for q in qubits],
             Expression([Symbol('begin'), *body]), env))
    stack.push(None)


def eval_begin(exp, env, stack):
    for x in exp[1:-1]:
        eval(x, env, stack)
        stack.pop()
    eval(exp[-1], env, stack)


def eval_let(exp, env, stack):
    (_, bindings, body) = exp
    let_env = Env(params=[], args=[], outer=env)
    for (var, val) in bindings:
        eval(val, env, stack)
        if not isinstance(var, Symbol):
            raise TypeError(f'var must be a symbol')
        let_env.assign(var.name, stack.pop())
    eval(body, let_env, stack)


def eval_letstar(exp, env, stack):
    (_, bindings, body) = exp
    letstar_env = Env(params=[], args=[], outer=env)
    for (var, val) in bindings:
        eval(val, letstar_env, stack)
        if not isinstance(var, Symbol):
            raise TypeError(f'var must be a symbol')
        letstar_env.assign(var.name, stack.pop())
    eval(body, letstar_env, stack)


def python_to_qlisp(exp):
    if isinstance(exp, (Expression, Channel)):
        return exp
    elif isinstance(exp, tuple):
        return Expression([python_to_qlisp(x) for x in exp])
    elif isinstance(exp, str):
        if len(exp) >= 2 and exp.startswith('"') and exp.endswith('"'):
            return exp[1:-1]
        else:
            return Symbol(exp)
    else:
        return exp


def get_ret(gen, env, stack):
    ret = yield from gen
    eval(python_to_qlisp(ret), env, stack)


def apply_gate(gate: Gate, args, env, stack):
    if gate.bindings is None and len(gate.params) > 0:
        bindings = dict(zip(gate.params, args))
        stack.push(
            Gate(gate.name, gate.params, gate.qubits, gate.body, gate.env,
                 bindings))
        return
    else:
        qubits = args

    if isinstance(gate.body, Expression):
        inner_env = Env(gate.qubits, qubits, gate.env)
        if gate.bindings is not None:
            inner_env.update(gate.bindings)
        eval(gate.body, inner_env, stack)
    else:
        pass


def apply(proc, args, env, stack):
    if isinstance(proc, Gate):
        apply_gate(proc, args, env, stack)
    elif isinstance(proc, Procedure):
        eval(proc.body, Env(proc.params, args, proc.env), stack)
    elif callable(proc):
        x = proc(*args)
        if inspect.isgenerator(x):
            for instruction in get_ret(x, env, stack):
                instruction = python_to_qlisp(instruction)
                if instruction[0].name.startswith('!'):
                    eval_instruction(instruction, env, stack)
                    # try:
                    #     cmd, target, *args = instruction
                    #     for a in reversed(args):
                    #         eval(a, env, stack)
                    #     args = [stack.pop() for _ in args]
                    #     if isinstance(target, Symbol):
                    #         target = target.name
                    #     execute(stack, cmd.name, target, *args)
                    # except:
                    #     raise Exception(f'bad instruction {instruction}')
                else:
                    eval(instruction, env, stack)
                    stack.pop()
        else:
            stack.push(x)
    else:
        raise TypeError('{} is not callable'.format(proc))


def eval_apply(exp, env, stack):
    (_, proc, args) = exp
    for exp in args[::-1]:
        eval(exp, env, stack)
    eval(proc, env, stack)
    proc = stack.pop()
    args = [stack.pop() for _ in args]
    apply(proc, args, env, stack)


def eval_channel(exp, env, stack):
    (_, *qubits, name) = exp
    if isinstance(name, Symbol):
        name = name.name
    qubits = tuple(q.name for q in qubits)
    stack.push(Channel(name, qubits))


def eval_instruction(instruction, env, stack):
    env = Env(params=[], args=[], outer=env)
    try:
        cmd, target, *args = instruction
        for a in reversed(args):
            eval(a, env, stack)
        args = [stack.pop() for _ in args]
        eval(target, env, stack)
        target = stack.pop()
        execute(stack, cmd.name, target, *args)
    except:
        raise Exception(f'bad instruction {instruction}')


__eval_table = {
    'quote': eval_quote,
    'if': eval_if,
    'cond': eval_cond,
    'while': eval_while,
    'define': eval_define,
    'defun': eval_defun,
    'gate': eval_gate,
    'set!': eval_set,
    'setq': eval_setq,
    'lambda': eval_lambda,
    'begin': eval_begin,
    'let': eval_let,
    'let*': eval_letstar,
    'apply': eval_apply,
    'channel': eval_channel,
}


def eval(x, env, stack):
    "Evaluate an expression in an environment."
    if isinstance(x, Symbol):  # variable reference
        stack.push(env.lookup(x.name))
    elif isinstance(x, Expression):
        if len(x) == 0:
            stack.push(None)
        elif isinstance(x[0], Symbol) and x[0].name.startswith('!'):
            eval_instruction(x, env, stack)
        elif isinstance(x[0], Symbol) and x[0].name in __eval_table:
            __eval_table[x[0].name](x, env, stack)
        else:
            eval(Expression([Symbol('apply'), x[0], x[1:]]), env, stack)
    else:  # constant literal
        stack.push(x)


################ virtual machine


def execute(stack, cmd, target, *args):
    if cmd == '!nop':
        pass
    elif cmd == '!set_waveform':
        stack.raw_waveforms[target] = args[0]
    elif cmd == '!add_waveform':
        stack.raw_waveforms[target] += args[0]
    elif cmd == '!set_phase':
        stack.phases[target] = args[0]
    else:
        pass
    stack.push(None)


################ Interaction: A REPL


def lispstr(exp):
    "Convert a Python object back into a Lisp-readable string."
    if isinstance(exp, Expression):
        return '(' + ' '.join(map(lispstr, exp)) + ')'
    else:
        return str(exp)


def repl(prompt='lis.py> '):
    "A prompt-read-eval-print loop."
    global_stack = Stack()
    global_env = standard_env()
    while True:
        eval(parse(input(prompt)), global_env, global_stack)
        val = global_stack.pop()
        if val is not None:
            print(lispstr(val))


if __name__ == "__main__":
    repl()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Node for an OPENQASM file identifier/version statement."""

import re
import warnings

from .node import Node


class Format(Node):
    """Node for an OPENQASM file identifier/version statement."""

    def __init__(self, value):
        """Create the version node."""
        super().__init__("format", None, None)
        parts = re.match(r'(\w+)\s+(\d+)\.(\d+)', value)
        self.language = parts.group(1)
        self.majorversion = parts.group(2)
        self.minorversion = parts.group(3)

    def version(self):
        """Return the version."""
        return "%s.%s" % (self.majorversion, self.minorversion)

    def qasm(self, prec=None):
        """Return the corresponding format string."""
        if prec is not None:
            warnings.warn('Parameter \'Format.qasm(..., prec)\' is no longer used and is being '
                          'deprecated.', DeprecationWarning, 2)
        return "%s %s;" % (self.language, self.version())


from typing import Any, Callable, NamedTuple

from waveforms.dicttree import flattenDict
from waveforms.qlisp.commands import (COMMAND, FREE, PUSH, READ, SYNC, TRIG,
                                      WRITE, CommandList, DataMap, RawData,
                                      Result)

from ..qlisp import MeasurementTask, QLispCode

from waveforms.qlisp.arch import Architecture, general_architecture



import sqlite3
from pathlib import Path


def _elms(seq):
    base = {'I'}
    for g1, g2 in zip(*seq):
        if (g1, g2) in [('CZ', 'CZ'), ('iSWAP', 'iSWAP'),
                        ('SQiSWAP', 'SQiSWAP')]:
            base.add(g1)
        elif (g1, g2) in [('C', 'Z'), ('Z', 'C')]:
            base.add('CZ')
        elif g1 == 'C' or g2 == 'C':
            base.add(g1 + g2)
        else:
            base.add(g1)
            base.add(g2)
    return base


class CliffordSequencyDB():
    def __init__(self, db=':memory:'):
        if db != ':memory:':
            self.db = Path(db)
        else:
            self.db = db
        self.conn = sqlite3.connect(':memory:')
        self.i2g = []
        self.g2i = {}
        if self.db == ':memory:' or not self.db.exists():
            self.createDatabase()
        else:
            source = sqlite3.connect(self.db)
            source.backup(self.conn)
        self.loadGateIndex()

    def flush(self):
        if self.db != ':memory:':
            self.conn.backup(sqlite3.connect(self.db))

    def close(self):
        self.flush()
        self.conn.close()

    def __del__(self):
        self.close()

    def createDatabase(self):
        cur = self.conn.cursor()
        # Create table
        cur.execute('''CREATE TABLE gates (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            name VARCHAR UNIQUE ON CONFLICT IGNORE
        )''')

        cur.execute('''CREATE TABLE sequences (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            depth INTEGER,
            clifford_index INTEGER CHECK(clifford_index >= 0 AND clifford_index < 11520),
            seq BLOB(128) UNIQUE ON CONFLICT IGNORE
        )''')

        cur.execute("CREATE INDEX clifford_index ON sequences(clifford_index)")
        cur.execute("CREATE INDEX seq_index ON sequences(seq)")

        cur.execute("INSERT INTO gates VALUES (0, 'I')")
        cur.executemany("INSERT INTO sequences VALUES (?, ?, ?, ?)",
                        [(0, 0, 0, b'')])
        gate_list = [
            'X',
            'Y',
            'Z',
            'X/2',
            '-X/2',
            'Y/2',
            '-Y/2',
            'S',
            '-S',
            'H',
            'CX',
            'XC',
            'CZ',
            'iSWAP',
            'T',
            '-T',
            'SQiSWAP',
        ]

        cur.executemany("INSERT INTO gates (name) VALUES (?)",
                        [(g, ) for g in gate_list])
        self.conn.commit()

    def loadGateIndex(self):
        cur = self.conn.cursor()
        self.i2g = [
            g for g, in cur.execute("SELECT name FROM gates ORDER BY id")
        ]
        self.g2i = {g: i for i, g in enumerate(self.i2g)}

    def updateGateIndex(self, gates):
        gates = set(gates) - set(self.i2g)
        if len(gates) > 0:
            cur = self.conn.cursor()
            cur.executemany("INSERT INTO gates (name) VALUES (?)",
                            [(g, ) for g in gates])
            self.conn.commit()
            self.loadGateIndex()

    def addSeq(self, seq, i):
        self.updateGateIndex(_elms(seq))
        buf, depth = self._seq2bytes(seq)
        self._insertSeq(depth, i, buf)

    def _insertSeq(self, depth, i, buf):
        cur = self.conn.cursor()
        cur.executemany(
            """INSERT INTO sequences
            (depth, clifford_index, seq) VALUES (?, ?, ?)""",
            [(depth, i, buf)])
        self.conn.commit()

    @staticmethod
    def _seqDepth(seq):
        A, B = seq
        if len(A) == 1 and A[0] == B[0] == 'I':
            return 0
        else:
            return len(A)

    def _seq2bytes(self, seq):
        buf = bytearray(128)
        depth = self._seqDepth(seq)
        if depth == 0:
            return buf, 0
        for i, (A, B) in enumerate(zip(*seq)):
            if A == 'C' or B == 'C':
                t = A + B
                A = B = t
            buf[2 * i] = self.g2i[A]
            buf[2 * i + 1] = self.g2i[B]
        return buf, depth

    def _bytes2seq(self, buf, depth):
        if depth == 0:
            return (('I', ), ('I', ))
        A, B = [], []
        for i in range(depth):
            A.append(self.i2g[buf[2 * i]])
            B.append(self.i2g[buf[2 * i + 1]])
        return tuple(A), tuple(B)

    def index2seq(self, i, lim=None):
        cur = self.conn.cursor()
        ret = []
        for buf, depth in cur.execute(f"""SELECT seq, depth FROM sequences
            INDEXED BY clifford_index WHERE clifford_index = {i}
            ORDER BY depth"""):
            ret.append(self._bytes2seq(buf, depth))
        return ret

    def seq2index(self, seq):
        try:
            buf, depth = self._seq2bytes(seq)
            cur = self.conn.cursor()
            q = cur.execute(
                f"""SELECT clifford_index FROM sequences
                    INDEXED BY seq_index
                    WHERE seq=:buf AND depth=:depth""", {
                    'buf': buf,
                    'depth': depth
                })
            return q.fetchone()[0]
        except:
            return None



from waveforms.namespace import NamespaceCache, NOTSET, DictDriver
import pytest
import copy


@pytest.fixture
def data():
    from config import config
    return copy.deepcopy(config)


def test_DictDriver(data):
    d = DictDriver(data)

    assert d.query('gates.rfUnitary.Q1.params.frequency') == 4675808085.0
    assert d.query('gates.rfUnitary.Q2.params.frequency') == 4354202438.483826
    assert d.query('gates.CZ') == data['gates']['CZ']
    assert set(d.keys('gates')) == {
        'Measure', 'Reset', 'rfUnitary', 'CZ', 'iSWAP', 'CR'
    }
    assert set(d.keys('*')) == {'__version__', 'station', 'chip', 'gates'}
    assert set(d.keys('gates.C*')) == {'CZ', 'CR'}
    d.delete_many(['gates.rfUnitary.Q1', 'gates.rfUnitary.Q2.params.duration'])
    assert data['gates']['rfUnitary'] == {
        "Q2": {
            "type": "default",
            "params": {
                "shape": "CosPulse",
                "frequency": 4354202438.483826,
                "amp": [[0, 1], [0, 0.658]],
                "phase": [[-1, 1], [-1, 1]],
                "DRAGScaling": 9.866314574173999e-10
            }
        }
    }
    assert d.query('gates.rfUnitary.Q1.params.frequency') == (
        NOTSET, 'gates.rfUnitary.Q1')


def test_DictDriver_update(data):
    d = DictDriver(data)

    d.update_many({
        'gates.rfUnitary.Q1.params.frequency': 4.6e9,
        'gates.rfUnitary.Q2.params.frequency': 4.3e9
    })

    assert d.query('gates.rfUnitary.Q1.params.frequency') == 4.6e9
    assert d.query('gates.rfUnitary.Q2.params.frequency') == 4.3e9

    d.update_many({
        'gates.rfUnitary.Q1.params': {
            "shape": "CosPulse",
            "frequency": 4675808085.0,
            "amp": [[0, 1], [0, 0.8204]],
            "duration": [[0, 1], [2e-08, 2e-08]],
            "phase": [[-1, 1], [-1, 1]],
            "DRAGScaling": 0.1
        },
        'gates.rfUnitary.Q1.params.frequency': 4.4e9
    })

    assert d.query('gates.rfUnitary.Q1.params.frequency') == 4.4e9
    assert d.query('gates.rfUnitary.Q1.params.DRAGScaling') == 0.1


def test_cache():
    log = NamespaceCache()
    log.cache('gate.rfUnitary.Q0',
              {'params': {
                  'frequency': 4.5e9,
                  'amp': 0.4,
                  'duration': 50e-9
              }})

    log.cache('gate.rfUnitary.Q1',
              {'params': {
                  'frequency': 4.6e9,
                  'amp': 0.41,
                  'duration': 52e-9
              }})

    log.cache(
        'gate.CZ', {
            'Q0_Q1': {
                'params': {
                    'amp': 0.25,
                    'duration': 50e-9,
                    'phi1': 0,
                    'phi2': 0
                }
            },
            'Q1_Q2': {
                'params': {
                    'amp': 0.15,
                    'duration': 50e-9,
                    'phi1': 0,
                    'phi2': 0
                }
            },
        })

    log.cache('gate.rfUnitary.Q4.params', (NOTSET, 'gate.rfUnitary.Q4'))

    assert log.query('gate.rfUnitary.Q0.params') == ({
        'frequency': 4.5e9,
        'amp': 0.4,
        'duration': 50e-9
    }, True)
    assert log.query('gate.rfUnitary.Q0.params.frequency') == (4.5e9, True)
    assert log.query('gate.rfUnitary') == ({
        'Q0': {
            'params': {
                'frequency': 4.5e9,
                'amp': 0.4,
                'duration': 50e-9
            }
        },
        'Q1': {
            'params': {
                'frequency': 4.6e9,
                'amp': 0.41,
                'duration': 52e-9
            }
        }
    }, False)
    assert log.query('gate.rfUnitary.Q3') == (None, False)
    assert log.keys('gate') is None
    assert log.keys('gate.rfUnitary') is None
    assert set(log.keys('gate.rfUnitary.Q0')) == {'params'}
    assert set(log.keys('gate.rfUnitary.Q0.params')) == {
        'frequency', 'amp', 'duration'
    }
    assert set(log.keys('gate.rfUnitary.Q0.params.frequency')) == set()
    assert set(log.keys('gate.CZ')) == {'Q0_Q1', 'Q1_Q2'}
    assert set(log.keys('gate.CZ.Q0*')) == {'Q0_Q1'}
    assert set(log.keys('gate.CZ.Q1*')) == {'Q1_Q2'}



from waveforms.security import verifyPassword, encryptPassword, InvalidKey


def verify(password, encrypted_password):
    try:
        verifyPassword(password, encrypted_password)
        return True
    except InvalidKey:
        return False


def test_verify():
    encrypted_password = encryptPassword("password")
    assert encrypted_password != "password"
    assert verify("password", encrypted_password)
    assert not verify("wrong password", encrypted_password)
